from __future__ import annotations

from typing import TYPE_CHECKING, List

import pandas as pd
import pytest

import great_expectations.exceptions as ge_exceptions
from great_expectations.datasource.fluent import SparkS3Datasource
from great_expectations.datasource.fluent.data_asset.path.path_data_asset import (
    PathDataAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.csv_asset import CSVAsset

if TYPE_CHECKING:
    from botocore.client import BaseClient


@pytest.fixture()
def aws_s3_bucket_name() -> str:
    return "test_bucket"


@pytest.fixture
def s3_bucket(s3_mock: BaseClient, aws_s3_bucket_name: str) -> str:
    bucket_name: str = aws_s3_bucket_name
    s3_mock.create_bucket(Bucket=bucket_name)
    return bucket_name


@pytest.fixture
def spark_s3_datasource(s3_mock, s3_bucket: str) -> SparkS3Datasource:
    test_df: pd.DataFrame = pd.DataFrame(data={"col1": [1, 2], "col2": [3, 4]})

    keys: List[str] = [
        "yellow_tripdata_sample_2024-01.csv",
        "yellow_tripdata_sample_2024-02.csv",
        "yellow_tripdata_sample_2024-03.csv",
        "yellow_tripdata_sample_2024-04.csv",
        "yellow_tripdata_sample_2024-05.csv",
        "yellow_tripdata_sample_2024-06.csv",
        "yellow_tripdata_sample_2024-07.csv",
        "yellow_tripdata_sample_2024-08.csv",
        "yellow_tripdata_sample_2024-09.csv",
        "yellow_tripdata_sample_2024-10.csv",
        "yellow_tripdata_sample_2024-11.csv",
        "yellow_tripdata_sample_2024-12.csv",
        "subfolder/for_recursive_search.csv",
    ]

    for key in keys:
        s3_mock.put_object(
            Bucket=s3_bucket,
            Body=test_df.to_csv(index=False).encode("utf-8"),
            Key=key,
        )

    return SparkS3Datasource(
        name="spark_s3_datasource",
        bucket=s3_bucket,
    )


@pytest.fixture
def csv_asset(spark_s3_datasource: SparkS3Datasource) -> PathDataAsset:
    asset = spark_s3_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
    )
    return asset


@pytest.mark.unit
def test_construct_spark_s3_datasource(spark_s3_datasource: SparkS3Datasource):
    assert spark_s3_datasource.name == "spark_s3_datasource"


@pytest.mark.unit
def test_add_csv_asset_to_datasource(spark_s3_datasource: SparkS3Datasource):
    asset_specified_metadata = {"asset_level_metadata": "my_metadata"}
    asset = spark_s3_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
        batch_metadata=asset_specified_metadata,
    )
    assert asset.name == "csv_asset"
    assert asset.batch_metadata == asset_specified_metadata


@pytest.mark.unit
def test_construct_csv_asset_directly():
    asset = CSVAsset(
        name="csv_asset",
    )
    assert asset.name == "csv_asset"


@pytest.mark.unit
def test_csv_asset_with_batching_regex_named_parameters(
    spark_s3_datasource: SparkS3Datasource,
):
    asset = spark_s3_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
    )
    batching_regex = r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
    batch_def = asset.add_batch_definition_monthly(name="batch def", regex=batching_regex)
    options = asset.get_batch_parameters_keys(partitioner=batch_def.partitioner)
    assert options == ("path", "year", "month")


@pytest.mark.unit
def test_csv_asset_with_non_string_batching_regex_named_parameters(
    spark_s3_datasource: SparkS3Datasource,
):
    asset = spark_s3_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
    )
    with pytest.raises(ge_exceptions.InvalidBatchRequestError):
        # price is an int which will raise an error
        asset.build_batch_request({"year": "2024", "month": 5})


@pytest.mark.unit
def test_add_csv_asset_with_recursive_file_discovery_to_datasource(
    spark_s3_datasource: SparkS3Datasource,
):
    """
    Tests that files from the subfolder(s) is returned
    when the s3_recursive_file_discovery-flag is set to True

    This makes the list_keys-function search and return files also
    from sub-directories on S3, not just the files in the folder
    specified with the s3_name_starts_with-parameter
    """
    asset_specified_metadata = {"asset_level_metadata": "my_metadata"}
    asset = spark_s3_datasource.add_csv_asset(
        name="csv_asset_recursive",
        header=True,
        infer_schema=True,
        batch_metadata=asset_specified_metadata,
        s3_recursive_file_discovery=True,
    )
    assert asset.batch_metadata == asset_specified_metadata



import pathlib
from logging import Logger
from typing import ClassVar, Literal, Optional, Type, Union

from great_expectations.compatibility.pyspark import (
    types as pyspark_types,
)
from great_expectations.datasource.fluent import BatchMetadata, _SparkFilePathDatasource
from great_expectations.datasource.fluent.data_asset.path.spark.csv_asset import (
    CSVAsset,
    DirectoryCSVAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.delta_asset import (
    DeltaAsset,
    DirectoryDeltaAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.json_asset import (
    DirectoryJSONAsset,
    JSONAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.orc_asset import (
    DirectoryORCAsset,
    ORCAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.parquet_asset import (
    DirectoryParquetAsset,
    ParquetAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.text_asset import (
    DirectoryTextAsset,
    TextAsset,
)
from great_expectations.datasource.fluent.data_connector import (
    FilesystemDataConnector,
)

logger: Logger

class SparkFilesystemDatasource(_SparkFilePathDatasource):
    # class attributes
    data_connector_type: ClassVar[Type[FilesystemDataConnector]] = ...

    # instance attributes
    type: Literal["spark_filesystem"] = "spark_filesystem"

    base_directory: pathlib.Path
    data_context_root_directory: Optional[pathlib.Path] = None
    def add_csv_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        glob_directive: str = "**/*",
        # vvv spark parameters for pyspark.sql.DataFrameReader.csv() (ordered as in pyspark v3.4.0)
        # path: PathOrPaths,
        # NA - path determined by asset
        # schema: Optional[Union[StructType, str]] = None,
        spark_schema: Optional[Union[pyspark_types.StructType, str]] = None,
        # sep: Optional[str] = None,
        sep: Optional[str] = None,
        # encoding: Optional[str] = None,
        encoding: Optional[str] = None,
        # quote: Optional[str] = None,
        quote: Optional[str] = None,
        # escape: Optional[str] = None,
        escape: Optional[str] = None,
        # comment: Optional[str] = None,
        comment: Optional[str] = None,
        # header: Optional[Union[bool, str]] = None,
        header: Optional[Union[bool, str]] = None,
        # inferSchema: Optional[Union[bool, str]] = None,
        infer_schema: Optional[Union[bool, str]] = None,
        # ignoreLeadingWhiteSpace: Optional[Union[bool, str]] = None,
        ignore_leading_white_space: Optional[Union[bool, str]] = None,
        # ignoreTrailingWhiteSpace: Optional[Union[bool, str]] = None,
        ignore_trailing_white_space: Optional[Union[bool, str]] = None,
        # nullValue: Optional[str] = None,
        null_value: Optional[str] = None,
        # nanValue: Optional[str] = None,
        nan_value: Optional[str] = None,
        # positiveInf: Optional[str] = None,
        positive_inf: Optional[str] = None,
        # negativeInf: Optional[str] = None,
        negative_inf: Optional[str] = None,
        # dateFormat: Optional[str] = None,
        date_format: Optional[str] = None,
        # timestampFormat: Optional[str] = None,
        timestamp_format: Optional[str] = None,
        # maxColumns: Optional[Union[int, str]] = None,
        max_columns: Optional[Union[int, str]] = None,
        # maxCharsPerColumn: Optional[Union[int, str]] = None,
        max_chars_per_column: Optional[Union[int, str]] = None,
        # maxMalformedLogPerPartition: Optional[Union[int, str]] = None,
        max_malformed_log_per_partition: Optional[Union[int, str]] = None,
        # mode: Optional[str] = None,
        mode: Optional[Literal["PERMISSIVE", "DROPMALFORMED", "FAILFAST"]] = None,
        # columnNameOfCorruptRecord: Optional[str] = None,
        column_name_of_corrupt_record: Optional[str] = None,
        # multiLine: Optional[Union[bool, str]] = None,
        multi_line: Optional[Union[bool, str]] = None,
        # charToEscapeQuoteEscaping: Optional[str] = None,
        char_to_escape_quote_escaping: Optional[str] = None,
        # samplingRatio: Optional[Union[float, str]] = None,
        sampling_ratio: Optional[Union[float, str]] = None,
        # enforceSchema: Optional[Union[bool, str]] = None,
        enforce_schema: Optional[Union[bool, str]] = None,
        # emptyValue: Optional[str] = None,
        empty_value: Optional[str] = None,
        # locale: Optional[str] = None,
        locale: Optional[str] = None,
        # lineSep: Optional[str] = None,
        line_sep: Optional[str] = None,
        # pathGlobFilter: Optional[Union[bool, str]] = None,
        path_glob_filter: Optional[Union[bool, str]] = None,
        # recursiveFileLookup: Optional[Union[bool, str]] = None,
        recursive_file_lookup: Optional[Union[bool, str]] = None,
        # modifiedBefore: Optional[Union[bool, str]] = None,
        modified_before: Optional[Union[bool, str]] = None,
        # modifiedAfter: Optional[Union[bool, str]] = None,
        modified_after: Optional[Union[bool, str]] = None,
        # unescapedQuoteHandling: Optional[str] = None,
        unescaped_quote_handling: Optional[
            Literal[
                "STOP_AT_CLOSING_QUOTE",
                "BACK_TO_DELIMITER",
                "STOP_AT_DELIMITER",
                "SKIP_VALUE",
                "RAISE_ERROR",
            ]
        ] = None,
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L309
        # Spark Generic File Reader Options vvv
        # ignore_corrupt_files: bool = ...,
        # ignore_missing_files: bool = ...,
        # Spark Generic File Reader Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-csv.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L604
        # CSV Specific Options vvv
        # prefer_date: Optional[bool] = None,
        # timestamp_ntz_format: Optional[str] = None,
        # enable_date_time_parsing_fallback: Optional[bool] = None,
        # CSV Specific Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
    ) -> CSVAsset: ...
    def add_directory_csv_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        glob_directive: str = "**/*",
        # Spark Directory Reader Options vvv
        data_directory: str | pathlib.Path = ...,
        # Spark Directory Reader Options ^^^
        # vvv spark parameters for pyspark.sql.DataFrameReader.csv() (ordered as in pyspark v3.4.0)
        # path: PathOrPaths,
        # NA - path determined by asset
        # schema: Optional[Union[StructType, str]] = None,
        spark_schema: Optional[Union[pyspark_types.StructType, str]] = None,
        # sep: Optional[str] = None,
        sep: Optional[str] = None,
        # encoding: Optional[str] = None,
        encoding: Optional[str] = None,
        # quote: Optional[str] = None,
        quote: Optional[str] = None,
        # escape: Optional[str] = None,
        escape: Optional[str] = None,
        # comment: Optional[str] = None,
        comment: Optional[str] = None,
        # header: Optional[Union[bool, str]] = None,
        header: Optional[Union[bool, str]] = None,
        # inferSchema: Optional[Union[bool, str]] = None,
        infer_schema: Optional[Union[bool, str]] = None,
        # ignoreLeadingWhiteSpace: Optional[Union[bool, str]] = None,
        ignore_leading_white_space: Optional[Union[bool, str]] = None,
        # ignoreTrailingWhiteSpace: Optional[Union[bool, str]] = None,
        ignore_trailing_white_space: Optional[Union[bool, str]] = None,
        # nullValue: Optional[str] = None,
        null_value: Optional[str] = None,
        # nanValue: Optional[str] = None,
        nan_value: Optional[str] = None,
        # positiveInf: Optional[str] = None,
        positive_inf: Optional[str] = None,
        # negativeInf: Optional[str] = None,
        negative_inf: Optional[str] = None,
        # dateFormat: Optional[str] = None,
        date_format: Optional[str] = None,
        # timestampFormat: Optional[str] = None,
        timestamp_format: Optional[str] = None,
        # maxColumns: Optional[Union[int, str]] = None,
        max_columns: Optional[Union[int, str]] = None,
        # maxCharsPerColumn: Optional[Union[int, str]] = None,
        max_chars_per_column: Optional[Union[int, str]] = None,
        # maxMalformedLogPerPartition: Optional[Union[int, str]] = None,
        max_malformed_log_per_partition: Optional[Union[int, str]] = None,
        # mode: Optional[str] = None,
        mode: Optional[Literal["PERMISSIVE", "DROPMALFORMED", "FAILFAST"]] = None,
        # columnNameOfCorruptRecord: Optional[str] = None,
        column_name_of_corrupt_record: Optional[str] = None,
        # multiLine: Optional[Union[bool, str]] = None,
        multi_line: Optional[Union[bool, str]] = None,
        # charToEscapeQuoteEscaping: Optional[str] = None,
        char_to_escape_quote_escaping: Optional[str] = None,
        # samplingRatio: Optional[Union[float, str]] = None,
        sampling_ratio: Optional[Union[float, str]] = None,
        # enforceSchema: Optional[Union[bool, str]] = None,
        enforce_schema: Optional[Union[bool, str]] = None,
        # emptyValue: Optional[str] = None,
        empty_value: Optional[str] = None,
        # locale: Optional[str] = None,
        locale: Optional[str] = None,
        # lineSep: Optional[str] = None,
        line_sep: Optional[str] = None,
        # pathGlobFilter: Optional[Union[bool, str]] = None,
        path_glob_filter: Optional[Union[bool, str]] = None,
        # recursiveFileLookup: Optional[Union[bool, str]] = None,
        recursive_file_lookup: Optional[Union[bool, str]] = None,
        # modifiedBefore: Optional[Union[bool, str]] = None,
        modified_before: Optional[Union[bool, str]] = None,
        # modifiedAfter: Optional[Union[bool, str]] = None,
        modified_after: Optional[Union[bool, str]] = None,
        # unescapedQuoteHandling: Optional[str] = None,
        unescaped_quote_handling: Optional[
            Literal[
                "STOP_AT_CLOSING_QUOTE",
                "BACK_TO_DELIMITER",
                "STOP_AT_DELIMITER",
                "SKIP_VALUE",
                "RAISE_ERROR",
            ]
        ] = None,
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L309
        # Spark Generic File Reader Options vvv
        # ignore_corrupt_files: bool = ...,
        # ignore_missing_files: bool = ...,
        # Spark Generic File Reader Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-csv.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L604
        # CSV Specific Options vvv
        # prefer_date: Optional[bool] = None,
        # timestamp_ntz_format: Optional[str] = None,
        # enable_date_time_parsing_fallback: Optional[bool] = None,
        # CSV Specific Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
    ) -> DirectoryCSVAsset: ...
    def add_parquet_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        glob_directive: str = "**/*",
        # Spark Generic File Reader Options vvv
        path_glob_filter: Optional[Union[bool, str]] = None,
        modified_before: Optional[Union[bool, str]] = None,
        modified_after: Optional[Union[bool, str]] = None,
        recursive_file_lookup: Optional[Union[bool, str]] = None,
        # Spark Generic File Reader Options ^^^
        # vvv spark parameters for pyspark.sql.DataFrameReader.parquet() (ordered as in pyspark v3.4.0)  # noqa: E501 # FIXME CoP
        # See https://spark.apache.org/docs/latest/sql-data-sources-parquet.html for more info.
        # Parquet Specific Options vvv
        merge_schema: Optional[Union[bool, str]] = None,
        datetime_rebase_mode: Optional[Literal["EXCEPTION", "CORRECTED", "LEGACY"]] = None,
        int_96_rebase_mode: Optional[Literal["EXCEPTION", "CORRECTED", "LEGACY"]] = None,
        # Parquet Specific Options ^^^
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L473
        # Spark Generic File Reader Options vvv
        # ignore_corrupt_files: bool = ...,
        # ignore_missing_files: bool = ...,
        # Spark Generic File Reader Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
    ) -> ParquetAsset: ...
    def add_directory_parquet_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        glob_directive: str = "**/*",
        # Spark Directory Reader Options vvv
        data_directory: str | pathlib.Path = ...,
        # Spark Directory Reader Options ^^^
        # Spark Generic File Reader Options vvv
        path_glob_filter: Optional[Union[bool, str]] = None,
        modified_before: Optional[Union[bool, str]] = None,
        modified_after: Optional[Union[bool, str]] = None,
        recursive_file_lookup: Optional[Union[bool, str]] = None,
        # Spark Generic File Reader Options ^^^
        # vvv spark parameters for pyspark.sql.DataFrameReader.parquet() (ordered as in pyspark v3.4.0)  # noqa: E501 # FIXME CoP
        # See https://spark.apache.org/docs/latest/sql-data-sources-parquet.html for more info.
        # Parquet Specific Options vvv
        merge_schema: Optional[Union[bool, str]] = None,
        datetime_rebase_mode: Optional[Literal["EXCEPTION", "CORRECTED", "LEGACY"]] = None,
        int_96_rebase_mode: Optional[Literal["EXCEPTION", "CORRECTED", "LEGACY"]] = None,
        # Parquet Specific Options ^^^
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L473
        # Spark Generic File Reader Options vvv
        # ignore_corrupt_files: bool = ...,
        # ignore_missing_files: bool = ...,
        # Spark Generic File Reader Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
    ) -> DirectoryParquetAsset: ...
    def add_orc_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        glob_directive: str = "**/*",
        # Spark Generic File Reader Options vvv
        path_glob_filter: Optional[Union[bool, str]] = None,
        modified_before: Optional[Union[bool, str]] = None,
        modified_after: Optional[Union[bool, str]] = None,
        recursive_file_lookup: Optional[Union[bool, str]] = None,
        # Spark Generic File Reader Options ^^^
        # ORC Specific Options vvv
        merge_schema: Optional[Union[bool, str]] = None,
        # ORC Specific Options ^^^
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L473
        # Spark Generic File Reader Options vvv
        # ignore_corrupt_files: bool = ...,
        # ignore_missing_files: bool = ...,
        # Spark Generic File Reader Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
    ) -> ORCAsset: ...
    def add_directory_orc_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        glob_directive: str = "**/*",
        # Spark Directory Reader Options vvv
        data_directory: str | pathlib.Path = ...,
        # Spark Directory Reader Options ^^^
        # Spark Generic File Reader Options vvv
        path_glob_filter: Optional[Union[bool, str]] = None,
        modified_before: Optional[Union[bool, str]] = None,
        modified_after: Optional[Union[bool, str]] = None,
        recursive_file_lookup: Optional[Union[bool, str]] = None,
        # Spark Generic File Reader Options ^^^
        # ORC Specific Options vvv
        merge_schema: Optional[Union[bool, str]] = None,
        # ORC Specific Options ^^^
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L473
        # Spark Generic File Reader Options vvv
        # ignore_corrupt_files: bool = ...,
        # ignore_missing_files: bool = ...,
        # Spark Generic File Reader Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
    ) -> DirectoryORCAsset: ...
    def add_json_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        glob_directive: str = "**/*",
        # vvv spark parameters for pyspark.sql.DataFrameReader.json() (ordered as in pyspark v3.4.0)
        # path: Union[str, List[str], RDD[str]],
        # NA - path determined by asset
        # schema: Optional[Union[StructType, str]] = None,
        spark_schema: Optional[Union[pyspark_types.StructType, str]] = None,
        # primitivesAsString: Optional[Union[bool, str]] = None,
        primitives_as_string: Optional[Union[bool, str]] = None,
        # prefersDecimal: Optional[Union[bool, str]] = None,
        prefers_decimal: Optional[Union[bool, str]] = None,
        # allowComments: Optional[Union[bool, str]] = None,
        allow_comments: Optional[Union[bool, str]] = None,
        # allowUnquotedFieldNames: Optional[Union[bool, str]] = None,
        allow_unquoted_field_names: Optional[Union[bool, str]] = None,
        # allowSingleQuotes: Optional[Union[bool, str]] = None,
        allow_single_quotes: Optional[Union[bool, str]] = None,
        # allowNumericLeadingZero: Optional[Union[bool, str]] = None,
        allow_numeric_leading_zero: Optional[Union[bool, str]] = None,
        # allowBackslashEscapingAnyCharacter: Optional[Union[bool, str]] = None,
        allow_backslash_escaping_any_character: Optional[Union[bool, str]] = None,
        # mode: Optional[str] = None,
        mode: Optional[Literal["PERMISSIVE", "DROPMALFORMED", "FAILFAST"]] = None,
        # columnNameOfCorruptRecord: Optional[str] = None,
        column_name_of_corrupt_record: Optional[str] = None,
        # dateFormat: Optional[str] = None,
        date_format: Optional[str] = None,
        # timestampFormat: Optional[str] = None,
        timestamp_format: Optional[str] = None,
        # multiLine: Optional[Union[bool, str]] = None,
        multi_line: Optional[Union[bool, str]] = None,
        # allowUnquotedControlChars: Optional[Union[bool, str]] = None,
        allow_unquoted_control_chars: Optional[Union[bool, str]] = None,
        # lineSep: Optional[str] = None,
        line_sep: Optional[str] = None,
        # samplingRatio: Optional[Union[float, str]] = None,
        sampling_ratio: Optional[Union[float, str]] = None,
        # dropFieldIfAllNull: Optional[Union[bool, str]] = None,
        drop_field_if_all_null: Optional[Union[bool, str]] = None,
        # encoding: Optional[str] = None,
        encoding: Optional[str] = None,
        # locale: Optional[str] = None,
        locale: Optional[str] = None,
        # pathGlobFilter: Optional[Union[bool, str]] = None,
        path_glob_filter: Optional[Union[bool, str]] = None,
        # recursiveFileLookup: Optional[Union[bool, str]] = None,
        recursive_file_lookup: Optional[Union[bool, str]] = None,
        # modifiedBefore: Optional[Union[bool, str]] = None,
        modified_before: Optional[Union[bool, str]] = None,
        # modifiedAfter: Optional[Union[bool, str]] = None,
        modified_after: Optional[Union[bool, str]] = None,
        # allowNonNumericNumbers: Optional[Union[bool, str]] = None,
        allow_non_numeric_numbers: Optional[Union[bool, str]] = None,
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L309
        # Spark Generic File Reader Options vvv
        # ignore_corrupt_files: bool = ...,
        # ignore_missing_files: bool = ...,
        # Spark Generic File Reader Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-json.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L309
        # JSON Specific Options vvv
        # timezone: str = ...,
        # timestamp_ntz_format: str = "yyyy-MM-dd'T'HH:mm:ss[.SSS]",
        # enable_date_time_parsing_fallback: bool = ...,
        # JSON Specific Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
    ) -> JSONAsset: ...
    def add_directory_json_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        glob_directive: str = "**/*",
        # Spark Directory Reader Options vvv
        data_directory: str | pathlib.Path = ...,
        # Spark Directory Reader Options ^^^
        # vvv spark parameters for pyspark.sql.DataFrameReader.json() (ordered as in pyspark v3.4.0)
        # path: Union[str, List[str], RDD[str]],
        # NA - path determined by asset
        # schema: Optional[Union[StructType, str]] = None,
        spark_schema: Optional[Union[pyspark_types.StructType, str]] = None,
        # primitivesAsString: Optional[Union[bool, str]] = None,
        primitives_as_string: Optional[Union[bool, str]] = None,
        # prefersDecimal: Optional[Union[bool, str]] = None,
        prefers_decimal: Optional[Union[bool, str]] = None,
        # allowComments: Optional[Union[bool, str]] = None,
        allow_comments: Optional[Union[bool, str]] = None,
        # allowUnquotedFieldNames: Optional[Union[bool, str]] = None,
        allow_unquoted_field_names: Optional[Union[bool, str]] = None,
        # allowSingleQuotes: Optional[Union[bool, str]] = None,
        allow_single_quotes: Optional[Union[bool, str]] = None,
        # allowNumericLeadingZero: Optional[Union[bool, str]] = None,
        allow_numeric_leading_zero: Optional[Union[bool, str]] = None,
        # allowBackslashEscapingAnyCharacter: Optional[Union[bool, str]] = None,
        allow_backslash_escaping_any_character: Optional[Union[bool, str]] = None,
        # mode: Optional[str] = None,
        mode: Optional[Literal["PERMISSIVE", "DROPMALFORMED", "FAILFAST"]] = None,
        # columnNameOfCorruptRecord: Optional[str] = None,
        column_name_of_corrupt_record: Optional[str] = None,
        # dateFormat: Optional[str] = None,
        date_format: Optional[str] = None,
        # timestampFormat: Optional[str] = None,
        timestamp_format: Optional[str] = None,
        # multiLine: Optional[Union[bool, str]] = None,
        multi_line: Optional[Union[bool, str]] = None,
        # allowUnquotedControlChars: Optional[Union[bool, str]] = None,
        allow_unquoted_control_chars: Optional[Union[bool, str]] = None,
        # lineSep: Optional[str] = None,
        line_sep: Optional[str] = None,
        # samplingRatio: Optional[Union[float, str]] = None,
        sampling_ratio: Optional[Union[float, str]] = None,
        # dropFieldIfAllNull: Optional[Union[bool, str]] = None,
        drop_field_if_all_null: Optional[Union[bool, str]] = None,
        # encoding: Optional[str] = None,
        encoding: Optional[str] = None,
        # locale: Optional[str] = None,
        locale: Optional[str] = None,
        # pathGlobFilter: Optional[Union[bool, str]] = None,
        path_glob_filter: Optional[Union[bool, str]] = None,
        # recursiveFileLookup: Optional[Union[bool, str]] = None,
        recursive_file_lookup: Optional[Union[bool, str]] = None,
        # modifiedBefore: Optional[Union[bool, str]] = None,
        modified_before: Optional[Union[bool, str]] = None,
        # modifiedAfter: Optional[Union[bool, str]] = None,
        modified_after: Optional[Union[bool, str]] = None,
        # allowNonNumericNumbers: Optional[Union[bool, str]] = None,
        allow_non_numeric_numbers: Optional[Union[bool, str]] = None,
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L309
        # Spark Generic File Reader Options vvv
        # ignore_corrupt_files: bool = ...,
        # ignore_missing_files: bool = ...,
        # Spark Generic File Reader Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-json.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L309
        # JSON Specific Options vvv
        # timezone: str = ...,
        # timestamp_ntz_format: str = "yyyy-MM-dd'T'HH:mm:ss[.SSS]",
        # enable_date_time_parsing_fallback: bool = ...,
        # JSON Specific Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
    ) -> DirectoryJSONAsset: ...
    def add_text_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        glob_directive: str = "**/*",
        # Spark Generic File Reader Options vvv
        path_glob_filter: Optional[Union[bool, str]] = None,
        modified_before: Optional[Union[bool, str]] = None,
        modified_after: Optional[Union[bool, str]] = None,
        recursive_file_lookup: Optional[Union[bool, str]] = None,
        # Spark Generic File Reader Options ^^^
        # Text Specific Options vvv
        # wholetext: bool = False,
        wholetext: bool = False,
        # lineSep: Optional[str] = None,
        line_sep: Optional[str] = None,
        # Text Specific Options ^^^
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L309
        # Spark Generic File Reader Options vvv
        # ignore_corrupt_files: bool = ...,
        # ignore_missing_files: bool = ...,
        # Spark Generic File Reader Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
    ) -> TextAsset: ...
    def add_directory_text_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        glob_directive: str = "**/*",
        # Spark Directory Reader Options vvv
        data_directory: str | pathlib.Path = ...,
        # Spark Directory Reader Options ^^^
        # Spark Generic File Reader Options vvv
        path_glob_filter: Optional[Union[bool, str]] = None,
        modified_before: Optional[Union[bool, str]] = None,
        modified_after: Optional[Union[bool, str]] = None,
        recursive_file_lookup: Optional[Union[bool, str]] = None,
        # Spark Generic File Reader Options ^^^
        # Text Specific Options vvv
        # wholetext: bool = False,
        wholetext: bool = False,
        # lineSep: Optional[str] = None,
        line_sep: Optional[str] = None,
        # Text Specific Options ^^^
        # vvv pyspark Docs <> Source Code mismatch
        # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html
        # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L309
        # Spark Generic File Reader Options vvv
        # ignore_corrupt_files: bool = ...,
        # ignore_missing_files: bool = ...,
        # Spark Generic File Reader Options ^^^
        # ^^^ pyspark Docs <> Source Code mismatch
    ) -> DirectoryTextAsset: ...
    def add_delta_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        glob_directive: str = "**/*",
        # Delta Specific Options vvv
        timestamp_as_of: Optional[str] = None,
        version_as_of: Optional[str] = None,
        # Delta Specific Options ^^^
    ) -> DeltaAsset: ...
    def add_delta_directory_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        glob_directive: str = "**/*",
        # Spark Directory Reader Options vvv
        data_directory: str | pathlib.Path = ...,
        # Spark Directory Reader Options ^^^
        # Delta Specific Options vvv
        timestamp_as_of: Optional[str] = None,
        version_as_of: Optional[str] = None,
        # Delta Specific Options ^^^
    ) -> DirectoryDeltaAsset: ...



from great_expectations.metrics.metric import NonEmptyString
from great_expectations.metrics.metric_results import MetricResult
from great_expectations.metrics.query import QueryMetric


class QueryRowCountResult(MetricResult[int]): ...


class QueryRowCount(QueryMetric[QueryRowCountResult]):
    name = "query.row_count"

    query: NonEmptyString



import pytest

from great_expectations.core.expectation_validation_result import (
    ExpectationSuiteValidationResult,
    ExpectationSuiteValidationResultMeta,
)
from great_expectations.data_context.types.resource_identifiers import ValidationResultIdentifier
from great_expectations.render.renderer import SlackRenderer


@pytest.mark.unit
def test_SlackRenderer_render(mocker):
    validation_result = ExpectationSuiteValidationResult(
        success=True,
        meta=ExpectationSuiteValidationResultMeta(
            **{
                "active_batch_definition": {
                    "batch_identifiers": {},
                    "data_asset_name": "taxi_data_1.csv",
                    "data_connector_name": "default_inferred_data_connector_name",
                    "datasource_name": "pandas",
                },
                "batch_markers": {
                    "ge_load_time": "20220727T154327.630107Z",
                    "pandas_data_fingerprint": "c4f929e6d4fab001fedc9e075bf4b612",
                },
                "batch_spec": {"path": "../data/taxi_data_1.csv"},
                "checkpoint_name": "single_validation_checkpoint",
                "expectation_suite_name": "taxi_suite_1",
                "great_expectations_version": "0.15.15",
                "run_id": {
                    "run_name": "20220727-114327-my-run-name-template",
                    "run_time": "2022-07-27T11:43:27.625252+00:00",
                },
                "validation_time": "20220727T154327.701100Z",
            }
        ),
        statistics={"successful_expectations": 3, "evaluated_expectations": 3},
        results=[],
        suite_name="my_suite",
    )
    key = mocker.MagicMock(spec=ValidationResultIdentifier)
    data_docs_pages = {key: {"local_site": "file:///localsite/index.html"}}
    notify_with = ["local_site"]

    slack_renderer = SlackRenderer()
    output = slack_renderer.render(
        validation_result=validation_result,
        data_docs_pages=data_docs_pages,
        notify_with=notify_with,
        validation_result_urls=["file:///localsite/index.html"],
    )

    assert output == [
        {
            "text": {
                "text": (
                    "\n*Asset*: `taxi_data_1.csv`  \n*Expectation Suite*: `my_suite`"
                    "\n*Summary*: *3* of *3* Expectations were met"
                ),
                "type": "mrkdwn",
            },
            "type": "section",
        },
        {
            "text": {
                "text": "*DataDocs* can be found here: `file:///localsite/index.html` \n (Please copy and paste link into a browser to view)\n",  # noqa: E501 # FIXME CoP
                "type": "mrkdwn",
            },
            "type": "section",
        },
    ]



from __future__ import annotations

import datetime
import json
import warnings
from copy import deepcopy
from typing import Dict, Optional, Union

from dateutil.parser import parse
from marshmallow import Schema, fields, post_load, pre_dump

from great_expectations._docs_decorators import public_api
from great_expectations.alias_types import JSONValues  # noqa: TC001 # FIXME CoP
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.data_context_key import DataContextKey


@public_api
class RunIdentifier(DataContextKey):
    """A RunIdentifier identifies a run (collection of validations) by run_name and run_time.

    Args:
        run_name: a string or None.
        run_time: a Datetime.datetime instance, a string, or None.
    """

    def __init__(
        self,
        run_name: Optional[str] = None,
        run_time: Optional[Union[datetime.datetime, str]] = None,
    ) -> None:
        super().__init__()
        assert run_name is None or isinstance(run_name, str), "run_name must be an instance of str"
        assert run_time is None or isinstance(run_time, (datetime.datetime, str)), (
            "run_time must be either None or an instance of str or datetime"
        )
        self._run_name = run_name

        if isinstance(run_time, str):
            try:
                run_time = parse(run_time)
            except (ValueError, TypeError):
                warnings.warn(
                    f'Unable to parse provided run_time str ("{run_time}") to datetime. Defaulting '
                    f"run_time to current time."
                )
                run_time = datetime.datetime.now(datetime.timezone.utc)

        if not run_time:
            try:
                run_time = parse(run_name)  # type: ignore[arg-type] # FIXME CoP
            except (ValueError, TypeError):
                run_time = None

        run_time = run_time or datetime.datetime.now(tz=datetime.timezone.utc)
        if not run_time.tzinfo:
            # This will change the timzeone to UTC, and convert the time based
            # on assuming that the current time is in local.
            run_time = run_time.astimezone(tz=datetime.timezone.utc)

        self._run_time = run_time

    @property
    def run_name(self):
        return self._run_name

    @property
    def run_time(self):
        return self._run_time

    def to_tuple(self):  # type: ignore[explicit-override] # FIXME
        return (
            self._run_name or "__none__",
            self._run_time.astimezone(tz=datetime.timezone.utc).strftime("%Y%m%dT%H%M%S.%fZ"),
        )

    def to_fixed_length_tuple(self):  # type: ignore[explicit-override] # FIXME
        return (
            self._run_name or "__none__",
            self._run_time.astimezone(tz=datetime.timezone.utc).strftime("%Y%m%dT%H%M%S.%fZ"),
        )

    def __repr__(self):  # type: ignore[explicit-override] # FIXME
        return json.dumps(self.to_json_dict())

    @override
    def __str__(self):
        return json.dumps(self.to_json_dict(), indent=2)

    @public_api
    def to_json_dict(self) -> Dict[str, JSONValues]:
        """Returns a JSON-serializable dict representation of this RunIdentifier.

        Returns:
            A JSON-serializable dict representation of this RunIdentifier.
        """
        myself = runIdentifierSchema.dump(self)
        return myself

    def set_run_time_tz(self, tz: datetime.timezone | None):
        """Localize the run_time to the given timezone, or default to system local tz.

        Args:
            tz: The timezone to localize to.
        """
        self._run_time = self._run_time.astimezone(tz=tz)

    @classmethod
    def from_tuple(cls, tuple_):
        return cls(tuple_[0], tuple_[1])

    @classmethod
    def from_fixed_length_tuple(cls, tuple_):
        return cls(tuple_[0], tuple_[1])


class RunIdentifierSchema(Schema):
    run_name = fields.Str()
    run_time = fields.AwareDateTime(format="iso", default_timezone=datetime.timezone.utc)

    @pre_dump
    def prepare_dump(self, data, **kwargs):
        data = deepcopy(data)
        data.set_run_time_tz(tz=None)  # sets to system local tz
        return data

    @post_load
    def make_run_identifier(self, data, **kwargs):
        return RunIdentifier(**data)


runIdentifierSchema = RunIdentifierSchema()



from __future__ import annotations

import base64
import logging
from pprint import pformat as pf
from sys import version_info as python_version
from typing import TYPE_CHECKING, Final, Sequence
from unittest.mock import ANY

import pytest
import sqlalchemy as sa
from pytest import param

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.snowflake import URL as SnowflakeURL
from great_expectations.compatibility.snowflake import snowflake
from great_expectations.data_context import AbstractDataContext
from great_expectations.datasource.fluent import (
    GxContextWarning,
    SQLDatasource,
    TestConnectionError,
)
from great_expectations.datasource.fluent.config_str import ConfigStr
from great_expectations.datasource.fluent.snowflake_datasource import (
    AccountIdentifier,
    SnowflakeDatasource,
    SnowflakeDsn,
)
from great_expectations.execution_engine import SqlAlchemyExecutionEngine

if TYPE_CHECKING:
    from pytest.mark.structures import ParameterSet  # type: ignore[import-not-found] # FIXME CoP
    from pytest_mock import MockerFixture

TEST_LOGGER: Final = logging.getLogger(__name__)

_EXAMPLE_PRIVATE_KEY: Final[bytes] = b"""-----BEGIN RSA PRIVATE KEY-----
MIICXAIBAAKBgHpFdCdKOGLaiMH9t1th1lKqJVcDwfnlP2lpneANbbsgHb6/4U2U
ua085zlNYhZ5xJsnSdqIAfragzuVYNk2OCpoN1Qkq4oWad0a4cEB2QBtP9js0dVW
xQObJM8t1ZLHB3Lw1NCqB6OefkP7XlE0w6aXRZ5IWwvVC86cBXVBmBXzAgMBAAEC
gYABR6TVnHNGpZ702OEIdde2ec12QbXQFdQ6GD7sz3cslEN7caq8Eyh2ZcLN2L+E
GLY0IY8mWHIc3BivkPq4i1a/JyRUzEToJvjVd8J1slrzz8ryMOAiPbxt33IpgGL3
/8KgOLYxjdg5bpn6sCZlOXy7WYjl1H8TBw8CzZF41Ha24QJBAM7U+8m0hyknbnBD
gKXGb0eHIBx0zlPaNJwDHUcJXujxbVfwVjKWLy07JoXRiAgPuVszIMhu0r+Xa87L
W2WLdTsCQQCXVm0He7SaytnrlAFck5/L4EjtWaAQGfmV4eawI2HemWMjj0tukdFt
wAWHDuKYMb+bg21OU2XQxollYYJfk/apAkBaSe10WuNZ2sXCKiWBuIMhZWJmKbNc
NXgb1tw0A2o0JBhIeDkYsij8BMNHTXWllz+iCUq5VG+ZhX9hcbJ/PIa7AkEAjfgd
v+9ktfGmDUGDJX23YmK9BywU5AX6BYkuB/6pSVFLl4hNkyRn+zUv+ksUdwH0Zccd
O2UxFnGpYtnenBsKQQJBAMX2tgFcg//t1Li4+dxlTvZZ/clZCLpWXp4HQgBwzxMN
wpDoF40OzNYrrKIboU4BJFOMOBWAS4DFDYGdfLVS99g=
-----END RSA PRIVATE KEY-----"""

_EXAMPLE_B64_ENCODED_PRIVATE_KEY: Final[bytes] = base64.standard_b64encode(_EXAMPLE_PRIVATE_KEY)


VALID_DS_CONFIG_PARAMS: Final[Sequence[ParameterSet]] = [
    param(
        {
            "connection_string": "snowflake://my_user:password@my_account/d_public/s_public?numpy=True&role=my_role&warehouse=my_wh"
        },
        id="connection_string str",
    ),
    param(
        {
            "connection_string": "snowflake://my_user:${MY_PASSWORD}@my_account/d_public/s_public?role=my_role&warehouse=my_wh"
        },
        id="connection_string ConfigStr - password sub",
    ),
    param(
        {
            "connection_string": "snowflake://${MY_USER}:${MY_PASSWORD}@my_account/d_public/s_public?role=my_role&warehouse=my_wh"
        },
        id="connection_string ConfigStr - user + password sub",
    ),
    param(
        {
            "connection_string": {
                "user": "my_user",
                "password": "password",
                "account": "my_account",
                "schema": "S_PUBLIC",
                "database": "D_PUBLIC",
                "role": "my_role",
                "warehouse": "my_wh",
            }
        },
        id="connection_string dict",
    ),
    param(
        {
            "connection_string": {
                "user": "my_user",
                "password": "${MY_PASSWORD}",
                "account": "my_account",
                "schema": "s_public",
                "database": "d_public",
                "role": "my_role",
                "warehouse": "my_wh",
            }
        },
        id="connection_string dict with password ConfigStr",
    ),
    param(
        {
            "connection_string": {
                "user": "my_user",
                "password": "DUMMY_VALUE",
                "account": "my_account",
                "database": "d_public",
                "schema": "s_public",
                "warehouse": "my_wh",
                "role": "my_role",
            },
            "kwargs": {"connect_args": {"private": _EXAMPLE_PRIVATE_KEY}},
        },
        id="private_key auth",
    ),
    param(
        {
            "connection_string": {
                "user": "my_user",
                "password": "DUMMY_VALUE",
                "account": "my_account",
                "database": "d_public",
                "schema": "s_public",
                "warehouse": "my_wh",
                "role": "my_role",
            },
            "kwargs": {"connect_args": {"private": _EXAMPLE_B64_ENCODED_PRIVATE_KEY}},
        },
        id="private_key auth b64 encoded",
    ),
]


@pytest.fixture
def seed_env_vars(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("MY_USER", "my_user")
    monkeypatch.setenv("MY_PASSWORD", "my_password")


@pytest.fixture
def sf_test_connection_noop(monkeypatch: pytest.MonkeyPatch) -> None:
    TEST_LOGGER.warning("Monkeypatching SnowflakeDatasource.test_connection() to a noop")

    def noop(self):
        TEST_LOGGER.info(".test_connection noop")

    monkeypatch.setattr(SnowflakeDatasource, "test_connection", noop)


@pytest.fixture
def sql_ds_test_connection_always_fail(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Monkey patch the parent class' test_connection() method to always fail.
    Useful for testing the extra handling that SnowflakeDatasource.test_connection() provides.
    """
    TEST_LOGGER.warning("Monkeypatching SQLDatasource.test_connection() to a always fail")

    def fail(self, test_assets: bool = True):
        TEST_LOGGER.info("SQLDatasource.test_connection() always fail")
        raise TestConnectionError("always fail")

    monkeypatch.setattr(SQLDatasource, "test_connection", fail)


@pytest.mark.unit
class TestAccountIdentifier:
    @pytest.mark.parametrize(
        "value",
        [
            "orgname.account_name",
            "orgname-account_name",
            "abc12345.us-east-1.aws",
            "xy12345.us-gov-west-1.aws",
            "xy12345.europe-west4.gcp",
            "xy12345.central-us.azure",
        ],
    )
    def test_str_and_repr_methods(self, value: str):
        account_identifier: AccountIdentifier = pydantic.parse_obj_as(AccountIdentifier, value)
        assert str(account_identifier) == value
        assert repr(account_identifier) == f"AccountIdentifier({value!r})"

    @pytest.mark.parametrize("account_name", ["account_name", "account-name"])
    def test_fmt1_parse(self, account_name: str):
        orgname = "orgname"
        value = f"{orgname}-{account_name}"
        print(f"{value=}")

        account_identifier = pydantic.parse_obj_as(AccountIdentifier, value)
        assert account_identifier.match

        assert account_identifier.account_name == account_name
        assert account_identifier.orgname == orgname
        assert account_identifier.as_tuple() == (orgname, account_name)

    @pytest.mark.parametrize(
        "value",
        [
            "abc12345.us-east-1.aws",
            "xy12345.us-gov-west-1.aws",
            "xy12345.europe-west4.gcp",
            "xy12345.central-us.azure",
        ],
    )
    def test_fmt2_parse(self, value: str):
        """
        The cloud portion is technically optional if the the provider is AWS, but expecting greatly
        simplifies our parsing logic.
        """
        print(f"{value=}")
        locator, _, _remainder = value.partition(".")
        cloud_region_id, _, cloud_service = _remainder.partition(".")

        account_identifier = pydantic.parse_obj_as(AccountIdentifier, value)
        assert account_identifier.match

        assert account_identifier.account_locator == locator
        assert account_identifier.region == (cloud_region_id or None)
        assert account_identifier.cloud == (cloud_service or None)

        assert account_identifier.as_tuple() == (
            locator,
            cloud_region_id,
            cloud_service,
        )

    @pytest.mark.parametrize(
        "value",
        [
            "foobar",
            "orgname.account-name",
            "orgname.account_name",
            "my_account.us-east-1",
            "xy12345.us-gov-west-1.aws.",
            "xy12345.europe-west4.gcp.bar",
            "xy12345.us-east-1.nope",
            "xy12345.",
            "xy12345.central-us.bazar",
            "xy12345_us-gov-west-1_aws",
            "xy12345_europe-west4_gcp",
            "xy12345_central-us_azure",
        ],
    )
    def test_invalid_formats(self, value: str):
        """
        Test that an invalid format that does not match but can still be stringified as
        the original value.
        """
        print(f"{value=}")
        account_identifier = pydantic.parse_obj_as(AccountIdentifier, value)
        assert not account_identifier.match
        assert str(account_identifier) == value


@pytest.mark.unit
def test_snowflake_dsn():
    dsn = pydantic.parse_obj_as(
        SnowflakeDsn,
        "snowflake://my_user:password@my_account/my_db/my_schema?role=my_role&warehouse=my_wh",
    )
    assert dsn.user == "my_user"
    assert dsn.password == "password"
    assert dsn.account_identifier == "my_account"
    assert dsn.database == "my_db"
    assert dsn.schema_ == "my_schema"
    assert dsn.role == "my_role"
    assert dsn.warehouse == "my_wh"


@pytest.mark.snowflake  # TODO: make this a unit test
@pytest.mark.parametrize(
    "config_kwargs",
    [
        *VALID_DS_CONFIG_PARAMS,
        param(
            {
                "user": "my_user",
                "password": "password",
                "account": "my_account",
                "schema": "s_public",
                "database": "d_public",
                "role": "my_role",
                "warehouse": "my_wh",
            },
            id="old config format - top level keys",
        ),
    ],
)
def test_valid_config(
    empty_file_context: AbstractDataContext,
    seed_env_vars: None,
    config_kwargs: dict,
    param_id: str,
):
    my_sf_ds_1 = SnowflakeDatasource(name=f"my_sf {param_id}", **config_kwargs)
    assert my_sf_ds_1

    my_sf_ds_1._data_context = empty_file_context  # attach to enable config substitution
    sql_engine = my_sf_ds_1.get_engine()
    assert isinstance(sql_engine, sa.engine.Engine)

    exec_engine = my_sf_ds_1.get_execution_engine()
    assert isinstance(exec_engine, SqlAlchemyExecutionEngine)


@pytest.mark.unit
@pytest.mark.parametrize(
    ["connection_string", "expected_errors"],
    [
        pytest.param(
            "${MY_CONFIG_VAR}",
            [
                {
                    "loc": ("connection_string", "__root__"),
                    "msg": "Only password, user may use config substitution;"
                    " 'domain' substitution not allowed",
                    "type": "value_error",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="illegal config substitution - full connection string",
        ),
        pytest.param(
            "snowflake://my_user:password@${MY_CONFIG_VAR}/db/schema",
            [
                {
                    "loc": ("connection_string", "__root__"),
                    "msg": "Only password, user may use config substitution;"
                    " 'domain' substitution not allowed",
                    "type": "value_error",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="illegal config substitution - account (domain)",
        ),
        pytest.param(
            "snowflake://my_user:password@account/${MY_CONFIG_VAR}/schema",
            [
                {
                    "loc": ("connection_string", "__root__"),
                    "msg": "Only password, user may use config substitution;"
                    " 'path' substitution not allowed",
                    "type": "value_error",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="illegal config substitution - database (path)",
        ),
        pytest.param(
            "snowflake://my_user:password@account/db/${MY_CONFIG_VAR}",
            [
                {
                    "loc": ("connection_string", "__root__"),
                    "msg": "Only password, user may use config substitution;"
                    " 'path' substitution not allowed",
                    "type": "value_error",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="illegal config substitution - schema (path)",
        ),
        pytest.param(
            "snowflake://my_user:password@my_account",
            [
                {
                    "loc": ("connection_string",),
                    "msg": "value is not a valid dict",
                    "type": "type_error.dict",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "ConfigStr - contains no config template strings in the format "
                    "'${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'",
                    "type": "value_error",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "URL path missing database/schema",
                    "type": "value_error.url.path",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="missing path",
        ),
        pytest.param(
            "snowflake://my_user:password@my_account//",
            [
                {
                    "loc": ("connection_string",),
                    "msg": "value is not a valid dict",
                    "type": "type_error.dict",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "ConfigStr - contains no config template strings in the format "
                    "'${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'",
                    "type": "value_error",
                },
                {
                    "ctx": {"msg": "missing database"},
                    "loc": ("connection_string",),
                    "msg": "URL path missing database/schema",
                    "type": "value_error.url.path",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="missing database + schema",
        ),
        pytest.param(
            "snowflake://my_user:password@my_account/my_db",
            [
                {
                    "loc": ("connection_string",),
                    "msg": "value is not a valid dict",
                    "type": "type_error.dict",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "ConfigStr - contains no config template strings in the format "
                    "'${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'",
                    "type": "value_error",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "URL path missing database/schema",
                    "type": "value_error.url.path",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="missing schema",
        ),
        pytest.param(
            "snowflake://my_user:password@my_account/my_db/",
            [
                {
                    "loc": ("connection_string",),
                    "msg": "value is not a valid dict",
                    "type": "type_error.dict",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "ConfigStr - contains no config template strings in the format "
                    "'${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'",
                    "type": "value_error",
                },
                {
                    "ctx": {"msg": "missing schema"},
                    "loc": ("connection_string",),
                    "msg": "URL path missing database/schema",
                    "type": "value_error.url.path",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="missing schema 2",
        ),
        pytest.param(
            "snowflake://my_user:password@my_account//my_schema",
            [
                {
                    "loc": ("connection_string",),
                    "msg": "value is not a valid dict",
                    "type": "type_error.dict",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "ConfigStr - contains no config template strings in the format "
                    "'${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'",
                    "type": "value_error",
                },
                {
                    "ctx": {"msg": "missing database"},
                    "loc": ("connection_string",),
                    "msg": "URL path missing database/schema",
                    "type": "value_error.url.path",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="missing database",
        ),
    ],
)
def test_missing_required_params(
    connection_string: str,
    expected_errors: list[dict],  # TODO: use pydantic error dict
):
    with pytest.raises(pydantic.ValidationError) as exc_info:
        ds = SnowflakeDatasource(
            name="my_sf_ds",
            connection_string=connection_string,
        )
        print(f"{ds!r}")

    print(f"\n\tErrors:\n{pf(exc_info.value.errors())}")
    assert exc_info.value.errors() == expected_errors


@pytest.mark.unit
@pytest.mark.parametrize(
    "connection_string, connect_args, expected_errors",
    [
        pytest.param(
            "snowflake://my_user:password@my_account/foo/bar?numpy=True",
            {
                "account": "my_account",
                "user": "my_user",
                "password": "123456",
                "schema": "foo",
                "database": "bar",
            },
            [
                {
                    "loc": ("__root__",),
                    "msg": "Provided both connection detail keyword args and `connection_string`.",
                    "type": "value_error",
                }
            ],
            id="both connection_string and connect_args",
        ),
        pytest.param(
            None,
            {},
            [
                {
                    "loc": ("connection_string",),
                    "msg": "none is not an allowed value",
                    "type": "type_error.none.not_allowed",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="neither connection_string nor connect_args",
        ),
        pytest.param(
            None,
            {
                "account": "my_account",
                "user": "my_user",
                "schema": "foo",
                "database": "bar",
                "role": "my_role",
                "warehouse": "my_wh",
            },
            [
                {
                    "loc": ("connection_string", "password"),
                    "msg": "field required",
                    "type": "value_error.missing",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "expected string or bytes-like object"
                    f"""{"" if python_version < (3, 11) else ", got 'dict'"}""",
                    "type": "type_error",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "str type expected",
                    "type": "type_error.str",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="incomplete connect_args",
        ),
        pytest.param(
            {
                "account": "my_account",
                "user": "my_user",
                "schema": "foo",
                "database": "bar",
                "warehouse": "baz",
                "role": "qux",
            },
            {},
            [
                {
                    "loc": ("connection_string", "password"),
                    "msg": "field required",
                    "type": "value_error.missing",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "expected string or bytes-like object"
                    f"""{"" if python_version < (3, 11) else ", got 'dict'"}""",
                    "type": "type_error",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "str type expected",
                    "type": "type_error.str",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="incomplete connection_string dict connect_args",
        ),
    ],
)
def test_conflicting_connection_string_and_args_raises_error(
    connection_string: ConfigStr | SnowflakeDsn | None | dict,
    connect_args: dict,
    expected_errors: list[dict],
):
    with pytest.raises(pydantic.ValidationError) as exc_info:
        _ = SnowflakeDatasource(
            name="my_sf_ds", connection_string=connection_string, **connect_args
        )
    assert exc_info.value.errors() == expected_errors


@pytest.mark.unit
@pytest.mark.parametrize(
    "connection_string, expected_errors",
    [
        pytest.param(
            "user_login_name:password@account_identifier/db/schema?role=my_role&warehouse=my_wh",
            [
                {
                    "loc": ("connection_string",),
                    "msg": "value is not a valid dict",
                    "type": "type_error.dict",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "ConfigStr - contains no config template strings in the format"
                    " '${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'",
                    "type": "value_error",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "invalid or missing URL scheme",
                    "type": "value_error.url.scheme",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="missing scheme",
        ),
        pytest.param(
            "snowflake://user_login_name@account_identifier/db/schema?role=my_role&warehouse=my_wh",
            [
                {
                    "loc": ("connection_string",),
                    "msg": "value is not a valid dict",
                    "type": "type_error.dict",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "ConfigStr - contains no config template strings in the format"
                    " '${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'",
                    "type": "value_error",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "URL password invalid",
                    "type": "value_error.url.password",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="bad password",
        ),
        pytest.param(
            "snowflake://user_login_name:password@/db/schema?role=my_role&warehouse=my_wh",
            [
                {
                    "loc": ("connection_string",),
                    "msg": "value is not a valid dict",
                    "type": "type_error.dict",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "ConfigStr - contains no config template strings in the format"
                    " '${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'",
                    "type": "value_error",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "URL domain invalid",
                    "type": "value_error.url.domain",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="bad domain",
        ),
        pytest.param(
            "snowflake://user_login_name:password@account_identifier/db/schema?warehouse=my_wh",
            [
                {
                    "ctx": {"msg": "missing role"},
                    "loc": ("connection_string",),
                    "msg": "URL query param missing",
                    "type": "value_error.url.query",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="missing role",
        ),
        pytest.param(
            "snowflake://user_login_name:password@account_identifier/db/schema?role=my_role",
            [
                {
                    "ctx": {"msg": "missing warehouse"},
                    "loc": ("connection_string",),
                    "msg": "URL query param missing",
                    "type": "value_error.url.query",
                },
                {
                    "loc": ("__root__",),
                    "msg": "Must provide either a connection string or a combination of account, "
                    "user, password, database, schema, warehouse, role as keyword args.",
                    "type": "value_error",
                },
            ],
            id="missing warehouse",
        ),
    ],
)
def test_invalid_connection_string_raises_dsn_error(
    connection_string: str, expected_errors: list[dict]
):
    with pytest.raises(pydantic.ValidationError) as exc_info:
        _ = SnowflakeDatasource(name="my_snowflake", connection_string=connection_string)

    assert expected_errors == exc_info.value.errors()


# TODO: Cleanup how we install test dependencies and remove this skipif
@pytest.mark.skipif(True if not snowflake else False, reason="snowflake is not installed")
@pytest.mark.unit
def test_get_execution_engine_succeeds():
    connection_string = (
        "snowflake://my_user:password@my_account/my_db/my_schema?role=my_role&warehouse=my_wh"
    )
    datasource = SnowflakeDatasource(name="my_snowflake", connection_string=connection_string)
    # testing that this doesn't raise an exception
    datasource.get_execution_engine()


@pytest.mark.snowflake
@pytest.mark.parametrize(
    ["config", "expected_called_with"],
    [
        param(
            {
                "name": "std connection_str",
                "connection_string": "snowflake://user:password@account/db/SCHEMA?warehouse=wh&role=role",
            },
            {"url": ANY},
            id="std connection_string str",
        ),
        param(
            {
                "name": "std connection_details",
                "connection_string": {
                    "user": "user",
                    "password": "password",
                    "account": "account",
                    "database": "db",
                    "schema": "schema",
                    "warehouse": "wh",
                    "role": "role",
                },
            },
            {
                "url": "snowflake://user:password@account/db/schema?application=great_expectations_core&role=role&warehouse=wh",
            },
            id="std connection_string dict",
        ),
        param(
            {
                "name": "conn str with connect_args",
                "connection_string": "snowflake://user:password@account/db/schema?warehouse=wh&role=role",
                "kwargs": {"connect_args": {"private_key": b"my_key"}},
            },
            {
                "connect_args": {"private_key": b"my_key"},
                "url": ANY,
            },
            id="connection_string str with connect_args",
        ),
        param(
            {
                "name": "conn details with connect_args",
                "connection_string": {
                    "user": "user",
                    "password": "password",
                    "account": "account",
                    "database": "db",
                    "schema": "schema",
                    "warehouse": "wh",
                    "role": "role",
                },
                "kwargs": {"connect_args": {"private_key": b"my_key"}},
            },
            {
                "connect_args": {"private_key": b"my_key"},
                "url": "snowflake://user:password@account/db/schema?application=great_expectations_core&role=role&warehouse=wh",
            },
            id="connection_string dict with connect_args",
        ),
    ],
)
def test_create_engine_is_called_with_expected_kwargs(
    mocker: MockerFixture,
    sf_test_connection_noop: None,
    ephemeral_context_with_defaults: AbstractDataContext,
    config: dict,
    expected_called_with: dict,
):
    create_engine_spy = mocker.spy(sa, "create_engine")

    datasource = ephemeral_context_with_defaults.data_sources.add_snowflake(**config)
    print(datasource)
    engine = datasource.get_engine()
    print(engine)

    create_engine_spy.assert_called_once_with(**expected_called_with)


@pytest.mark.snowflake
@pytest.mark.parametrize(
    ("password", "encoded_password"),
    [
        pytest.param("abc", "abc", id="no need to encode"),
        pytest.param("a@b", "a%40b", id="encode it"),
    ],
)
def test_test_connection_encoding(
    mocker: MockerFixture,
    ephemeral_context_with_defaults: AbstractDataContext,
    password: str,
    encoded_password: str,
):
    account = "account"
    user = "foo"
    role = "role"
    warehouse = "warehouse"
    db = "db"
    schema = "schema"
    create_engine_spy = mocker.patch.object(sa, "create_engine")

    datasource = ephemeral_context_with_defaults.data_sources.add_snowflake(
        name="foo",
        account=account,
        user=user,
        password=password,
        database=db,
        schema=schema,
        role=role,
        warehouse=warehouse,
    )
    datasource.get_engine()

    create_engine_spy.assert_called_with(
        url=SnowflakeURL(
            account=account,
            user=user,
            password=encoded_password,
            database=db,
            schema=schema,
            role=role,
            warehouse=warehouse,
            application="great_expectations_core",
        )
    )


@pytest.mark.unit
@pytest.mark.parametrize("ds_config", VALID_DS_CONFIG_PARAMS)
class TestConvenienceProperties:
    def test_schema(
        self,
        ds_config: dict,
        seed_env_vars: None,
        param_id: str,
        ephemeral_context_with_defaults: AbstractDataContext,
    ):
        datasource = SnowflakeDatasource(name=param_id, **ds_config)
        if isinstance(datasource.connection_string, ConfigStr):
            # expect a warning if connection string is a ConfigStr
            with pytest.warns(GxContextWarning):
                assert not datasource.schema_, (
                    "Don't expect schema to be available without config_provider"
                )
            # attach context to enable config substitution
            datasource._data_context = ephemeral_context_with_defaults
            _ = datasource.schema_
        else:
            assert datasource.schema_ == datasource.connection_string.schema_.lower()

    def test_database(
        self,
        ds_config: dict,
        seed_env_vars: None,
        param_id: str,
        ephemeral_context_with_defaults: AbstractDataContext,
    ):
        datasource = SnowflakeDatasource(name=param_id, **ds_config)
        if isinstance(datasource.connection_string, ConfigStr):
            # expect a warning if connection string is a ConfigStr
            with pytest.warns(GxContextWarning):
                assert not datasource.database, (
                    "Don't expect database to be available without config_provider"
                )
            # attach context to enable config substitution
            datasource._data_context = ephemeral_context_with_defaults
            _ = datasource.database
        else:
            assert datasource.database == datasource.connection_string.database

    def test_warehouse(
        self,
        ds_config: dict,
        seed_env_vars: None,
        param_id: str,
        ephemeral_context_with_defaults: AbstractDataContext,
    ):
        datasource = SnowflakeDatasource(name=param_id, **ds_config)
        if isinstance(datasource.connection_string, ConfigStr):
            # expect a warning if connection string is a ConfigStr
            with pytest.warns(GxContextWarning):
                assert not datasource.warehouse, (
                    "Don't expect warehouse to be available without config_provider"
                )
            # attach context to enable config substitution
            datasource._data_context = ephemeral_context_with_defaults
            _ = datasource.warehouse
        else:
            assert datasource.warehouse == datasource.connection_string.warehouse

    def test_role(
        self,
        ds_config: dict,
        seed_env_vars: None,
        param_id: str,
        ephemeral_context_with_defaults: AbstractDataContext,
    ):
        datasource = SnowflakeDatasource(name=param_id, **ds_config)
        if isinstance(datasource.connection_string, ConfigStr):
            # expect a warning if connection string is a ConfigStr
            with pytest.warns(GxContextWarning):
                assert not datasource.role, (
                    "Don't expect role to be available without config_provider"
                )
            # attach context to enable config substitution
            datasource._data_context = ephemeral_context_with_defaults
            _ = datasource.role
        else:
            assert datasource.role == datasource.connection_string.role

    def test_account(
        self,
        ds_config: dict,
        seed_env_vars: None,
        param_id: str,
        ephemeral_context_with_defaults: AbstractDataContext,
    ):
        datasource = SnowflakeDatasource(name=param_id, **ds_config)
        if isinstance(datasource.connection_string, ConfigStr):
            # expect a warning if connection string is a ConfigStr
            with pytest.warns(GxContextWarning):
                assert not datasource.account, (
                    "Don't expect account to be available without config_provider"
                )
            # attach context to enable config substitution
            datasource._data_context = ephemeral_context_with_defaults
            _ = datasource.account
        else:
            assert datasource.account == datasource.connection_string.account


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



from __future__ import annotations

import random
import string
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, field
from functools import cached_property
from typing import TYPE_CHECKING, Generator, Generic, Hashable, Mapping, Optional, TypeVar
from uuid import UUID, uuid4

import pandas as pd

from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context.data_context.abstract_data_context import AbstractDataContext
from great_expectations.datasource.fluent.interfaces import Batch, DataAsset

if TYPE_CHECKING:
    import pytest
    from pytest import FixtureRequest

    from tests.integration.test_utils.data_source_config.sql import SessionSQLEngineManager


_ColumnTypes = TypeVar("_ColumnTypes")


@dataclass(frozen=True)
class DataSourceTestConfig(ABC, Generic[_ColumnTypes]):
    name: Optional[str] = None
    table_name: Optional[str] = None  # Overrides random table name generation
    column_types: Optional[Mapping[str, _ColumnTypes]] = None
    extra_column_types: Mapping[str, Mapping[str, _ColumnTypes]] = field(default_factory=dict)

    @property
    @abstractmethod
    def label(self) -> str:
        """Label that will show up in test name."""
        ...

    @property
    @abstractmethod
    def pytest_mark(self) -> pytest.MarkDecorator:
        """Mark for pytest"""
        ...

    @abstractmethod
    def create_batch_setup(
        self,
        request: FixtureRequest,
        data: pd.DataFrame,
        extra_data: Mapping[str, pd.DataFrame],
        context: AbstractDataContext,
        # This violates the interface segration principle (the I in SOLID) since we now make
        # non-SQL datasources rely on an argument that only SQL datasources are need.
        # However, this is simpler than adding an additional layer to decouple this interface.
        # If the SQL and non-SQL test interfaces diverge more significantly we should consider
        # refactoring these tests.
        # One possible fix is to remove this method from this class and create a sql and
        # non-sql subclass. We'd like need to update _ConfigT to be bounded by a union of
        # these subclasses and update callers of create_batch_setup.
        engine_manager: Optional[SessionSQLEngineManager] = None,
    ) -> BatchTestSetup:
        """Create a batch setup object for this data source."""

    @property
    def test_id(self) -> str:
        parts: list[Optional[str]] = [self.label, self.name]
        non_null_parts = [p for p in parts if p is not None]

        return "-".join(non_null_parts)

    @override
    def __eq__(self, value: object) -> bool:
        if not isinstance(value, DataSourceTestConfig):
            return False
        return all(
            [
                super().__eq__(value),
                self.label == value.label,
                self.pytest_mark == value.pytest_mark,
            ]
        )

    @override
    def __hash__(self) -> int:
        hashable_col_types = dict_to_tuple(self.column_types) if self.column_types else None
        hashable_extra_col_types = dict_to_tuple(
            {k: dict_to_tuple(self.extra_column_types[k]) for k in sorted(self.extra_column_types)}
        )
        return hash(
            (
                self.__class__.name,
                self.test_id,
                hashable_col_types,
                hashable_extra_col_types,
            )
        )


_ConfigT = TypeVar("_ConfigT", bound=DataSourceTestConfig)
_AssetT = TypeVar("_AssetT", bound=DataAsset)


class BatchTestSetup(ABC, Generic[_ConfigT, _AssetT]):
    """ABC for classes that set up and tear down batches."""

    def __init__(self, config: _ConfigT, data: pd.DataFrame, context: AbstractDataContext) -> None:
        self.config = config
        self.data = data
        self.context = context

    @abstractmethod
    def make_asset(self) -> _AssetT: ...

    @abstractmethod
    def make_batch(self) -> Batch: ...

    @abstractmethod
    def setup(self) -> None: ...

    @abstractmethod
    def teardown(self) -> None: ...

    @contextmanager
    def data_context_test_context(self) -> Generator[AbstractDataContext, None, None]:
        """Receive a DataContext and ensure proper setup and teardown regardless of errors."""
        try:
            self.setup()
            yield self.context
        finally:
            self.teardown()

    @contextmanager
    def asset_test_context(self) -> Generator[_AssetT, None, None]:
        """Receive an Asset and ensure proper setup and teardown regardless of errors."""
        try:
            self.setup()
            yield self.make_asset()
        finally:
            self.teardown()

    @contextmanager
    def batch_test_context(self) -> Generator[Batch, None, None]:
        """Receive a Batch and ensure proper setup and teardown regardless of errors."""
        try:
            self.setup()
            yield self.make_batch()
        finally:
            self.teardown()

    @staticmethod
    def _random_resource_name() -> str:
        return "".join(random.choices(string.ascii_lowercase, k=10))

    @cached_property
    def id(self) -> UUID:
        return uuid4()


def dict_to_tuple(d: Mapping[str, Hashable]) -> tuple[tuple[str, Hashable], ...]:
    return tuple((key, d[key]) for key in sorted(d))


def hash_data_frame(df: pd.DataFrame) -> int:
    return hash(tuple(pd.util.hash_pandas_object(df).array))



import uuid
from typing import Any

import pandas as pd
import pytest

import great_expectations as gx
import great_expectations.compatibility.postgresql as postgresql_dialect
import great_expectations.expectations as gxe
from great_expectations.expectations.metrics.util import MAX_RESULT_RECORDS
from great_expectations.render.components import (
    AtomicDiagnosticRendererType,
    RenderedAtomicContent,
    RenderedAtomicValue,
)
from great_expectations.render.renderer.observed_value_renderer import ObservedValueRenderState
from great_expectations.render.renderer_configuration import (
    MetaNotesFormat,
    RendererSchema,
    RendererTableValue,
    RendererValueType,
)
from tests.integration.conftest import (
    MultiSourceBatch,
    MultiSourceTestConfig,
    multi_source_batch_setup,
)
from tests.integration.data_sources_and_expectations.data_sources.test_comparison_to_base import (
    ALL_COMPARISON_TO_BASE_SOURCES,
)
from tests.integration.test_utils.data_source_config import SqliteDatasourceTestConfig
from tests.integration.test_utils.data_source_config.postgres import (
    PostgresBatchTestSetup,
    PostgreSQLDatasourceTestConfig,
)

SQLITE_ONLY = [
    MultiSourceTestConfig(
        comparison=SqliteDatasourceTestConfig(),
        base=SqliteDatasourceTestConfig(),
    )
]

COMPARISON_DATA = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})

BASE_DATA = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6], "c": [4, 5, 6]})


@pytest.mark.parametrize(
    "base_query,comparison_query",
    [
        pytest.param(
            "SELECT a, b FROM {batch} ORDER BY a, b",
            "SELECT a, b FROM {source_table} ORDER BY a, b",
            id="multiple_columns_multiple_rows",
        ),
        pytest.param(
            "SELECT a FROM {batch} ORDER BY a",
            "SELECT a FROM {source_table} ORDER BY a",
            id="one_column_multiple_rows",
        ),
        pytest.param(
            "SELECT a, b FROM {batch} ORDER BY b LIMIT 1",
            "SELECT a, b FROM {source_table} ORDER BY b LIMIT 1",
            id="multiple_columns_one_row",
        ),
        pytest.param(
            "SELECT a, b FROM {batch} LIMIT 0",
            "SELECT a, b FROM {source_table} LIMIT 0",
            id="both_results_are_empty",
        ),
        pytest.param(
            "SELECT a, c FROM {batch} ORDER BY c",
            "SELECT a, b FROM {source_table} ORDER BY b",
            id="column_names_different_values_the_same",
        ),
    ],
)
@multi_source_batch_setup(
    multi_source_test_configs=ALL_COMPARISON_TO_BASE_SOURCES,
    base_data=BASE_DATA,
    comparison_data=COMPARISON_DATA,
)
def test_expect_query_results_to_match_comparison_success(
    multi_source_batch: MultiSourceBatch, base_query: str, comparison_query: str
):
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            base_query=base_query,
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=comparison_query.replace(
                "{source_table}", multi_source_batch.comparison_table_name
            ),
        )
    )
    assert result.success


@pytest.mark.parametrize(
    "base_query,comparison_query",
    [
        pytest.param(
            "SELECT * FROM {batch}",
            "SELECT * FROM {source_table}",
            id="duplicate_values_across_rows",
        ),
        pytest.param(
            "SELECT a, b FROM {batch} LIMIT 2",
            "SELECT a, b FROM {source_table}",
            id="row_count_mismatch",
        ),
        pytest.param(
            "SELECT a FROM {batch} ORDER BY a",
            "SELECT b FROM {source_table} ORDER BY a",
            id="column_value_mismatch",
        ),
        pytest.param(
            "SELECT * FROM {batch} LIMIT 0",
            "SELECT * FROM {source_table} ORDER BY a",
            id="one_result_is_empty",
        ),
    ],
)
@multi_source_batch_setup(
    multi_source_test_configs=ALL_COMPARISON_TO_BASE_SOURCES,
    base_data=BASE_DATA,
    comparison_data=COMPARISON_DATA,
)
def test_expect_query_results_to_match_comparison_failure(
    multi_source_batch: MultiSourceBatch, base_query: str, comparison_query: str
):
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            base_query=base_query,
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=comparison_query.replace(
                "{source_table}", multi_source_batch.comparison_table_name
            ),
        )
    )
    assert not result.success
    assert not result.exception_info["raised_exception"]


@pytest.mark.parametrize(
    "mostly,success",
    [
        pytest.param(0.9, False, id="mostly_failure"),
        pytest.param(0.5, True, id="mostly_success"),
    ],
)
@multi_source_batch_setup(
    multi_source_test_configs=SQLITE_ONLY,
    base_data=BASE_DATA,
    comparison_data=COMPARISON_DATA,
)
def test_expect_query_results_to_match_comparison_mostly(
    multi_source_batch: MultiSourceBatch, mostly: float, success: bool
):
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            base_query="SELECT a, b FROM {batch} LIMIT 2",
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=f"SELECT a, b FROM {multi_source_batch.comparison_table_name}",
            mostly=mostly,
        )
    )
    assert result.success is success


MAX_LENGTH_BASE_DATA = pd.DataFrame(
    {
        "a": list(range(100, 300)),
        "b": list(range(100, 200)) + ([None] * 100),
        "c": list(range(200, 300)) + ([None] * 100),
        "no_dups": [1, 2, 3] + ([None] * 197),
        "has_dups": [1, 1, 2, 3] + ([None] * 196),
    }
)

MAX_LENGTH_COMPARISON_DATA = pd.DataFrame(
    {
        "a": list(range(0, 200)),
        "b": list(range(100, 200)) + ([None] * 100),
        "high_numbers": list(range(1000, 1200)),
        "has_dups": [1, 1, 2, 3] + ([None] * 196),
    }
)


@pytest.mark.parametrize(
    "base_query,comparison_query,unexpected_percent,unexpected_count",
    [
        pytest.param(
            "SELECT b FROM {batch} ORDER BY b",  # 100 records
            "SELECT b FROM {source_table} ORDER BY b",  # 100 records
            0,
            0,
            id="only_match",
        ),
        pytest.param(
            "SELECT a FROM {batch}",  # 200 records (half match)
            "SELECT b FROM {source_table}",  # 100 records
            50,
            100,
            id="match_and_unexpected",
        ),
        pytest.param(
            "SELECT b FROM {batch} ORDER BY b",  # 100 records
            "SELECT a FROM {source_table}",  # 200 records (half match)
            50,
            100,
            id="match_and_missing",
        ),
        pytest.param(
            "SELECT a FROM {batch}",  # 200 low numbers
            "SELECT high_numbers FROM {source_table}",  # 200 high numbers
            100,
            200,
            id="missing_and_unexpected",
        ),
        pytest.param(
            "SELECT a FROM {batch}",  # 200 records (half match)
            "SELECT a FROM {source_table}",  # 200 records
            50,
            100,
            id="match_and_missing_and_unexpected",
        ),
        pytest.param(
            "SELECT a FROM {batch}",  # 200 records
            "SELECT b FROM {source_table}",  # 100 different records
            50,
            100,
            id="only_unexpected",
        ),
        pytest.param(
            "SELECT c FROM {batch} ORDER BY c",  # 100 records
            "SELECT * FROM {source_table} LIMIT 0",  # 0 records
            100,
            200,
            id="only_missing",
        ),
        pytest.param(
            "SELECT * FROM {batch} LIMIT 0",  # 0 records
            "SELECT * FROM {source_table} LIMIT 0",  # 0 records
            0,
            0,
            id="nothing_to_compare",
        ),
        pytest.param(
            "SELECT has_dups FROM {batch}",  # 4 records (2 are dups)
            "SELECT has_dups FROM {source_table}",  # same 4 records
            0,
            0,
            id="has_dups_success",
        ),
        pytest.param(
            "SELECT no_dups FROM {batch}",  # 3 records (no dups)
            "SELECT has_dups FROM {source_table}",  # same 3 records + 1 dup
            1 / 200 * 100,
            1,
            id="has_dups_failure",
        ),
    ],
)
@multi_source_batch_setup(
    multi_source_test_configs=SQLITE_ONLY,
    base_data=MAX_LENGTH_BASE_DATA,
    comparison_data=MAX_LENGTH_COMPARISON_DATA,
)
def test_expect_query_results_to_match_comparison_unexpected_percent(
    multi_source_batch: MultiSourceBatch,
    base_query: str,
    comparison_query: str,
    unexpected_percent: float,
    unexpected_count: int,
):
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            base_query=base_query,
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=comparison_query.replace(
                "{source_table}", multi_source_batch.comparison_table_name
            ),
        )
    )
    assert result.result["unexpected_percent"] == pytest.approx(unexpected_percent)
    assert result.result["unexpected_count"] == unexpected_count


MISSING_AND_UNEXPECTED_DF = pd.DataFrame(
    {
        "id": [1, 1, 1, 2, 2, 3],
        "incorrect_id": [1, 1, 1, 2, 2, 4],
        "source": list("AAABBC"),
        "all_matches": list("CBBAAA"),
        "all_matches_reversed": list("CBBAAA"),
        "missing_and_unexpected": list("AAAAAD"),
    }
)


@pytest.mark.parametrize(
    ("base_query", "missing_rows", "unexpected_rows"),
    [
        pytest.param(
            "SELECT all_matches FROM {batch}",
            [],
            [],
            id="all_match",
        ),
        pytest.param(
            "SELECT all_matches_reversed FROM {batch}",
            [],
            [],
            id="all_match_order_agnostic",
        ),
        pytest.param(
            "SELECT missing_and_unexpected FROM {batch}",
            [
                {"source": "B"},
                {"source": "B"},
                {"source": "C"},
            ],
            [
                {"missing_and_unexpected": "A"},
                {"missing_and_unexpected": "A"},
                {"missing_and_unexpected": "D"},
            ],
            id="some_matches_missing_and_unexpected",
        ),
    ],
)
@multi_source_batch_setup(
    multi_source_test_configs=SQLITE_ONLY,
    base_data=MISSING_AND_UNEXPECTED_DF,
    comparison_data=MISSING_AND_UNEXPECTED_DF,
)
def test_expect_query_results_to_match_comparison_missing_and_unexpected_values(
    multi_source_batch: MultiSourceBatch,
    base_query: str,
    missing_rows: list[dict[str, Any]],
    unexpected_rows: list[dict[str, Any]],
) -> None:
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            base_query=base_query,
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=f"SELECT source FROM {multi_source_batch.comparison_table_name}",
        )
    )

    assert result.result["details"] == {
        "unexpected_rows": unexpected_rows,
        "missing_rows": missing_rows,
    }


@pytest.mark.parametrize(
    ("base_query", "comparison_query", "missing_rows", "unexpected_rows"),
    [
        pytest.param(
            "SELECT incorrect_id, source FROM {batch}",
            "SELECT id, source FROM {source_table}",
            [{"source": "C", "id": 3}],
            [{"source": "C", "incorrect_id": 4}],
            id="One bad row, but others match despite col names",
        ),
        pytest.param(
            "SELECT id, source FROM {batch}",
            "SELECT source, id FROM {source_table}",
            [
                {"source": "A", "id": 1},
                {"source": "A", "id": 1},
                {"source": "A", "id": 1},
                {"source": "B", "id": 2},
                {"source": "B", "id": 2},
                {"source": "C", "id": 3},
            ],
            [
                {"source": "A", "id": 1},
                {"source": "A", "id": 1},
                {"source": "A", "id": 1},
                {"source": "B", "id": 2},
                {"source": "B", "id": 2},
                {"source": "C", "id": 3},
            ],
            id="Same data, but cols in the wrong order",
        ),
    ],
)
@multi_source_batch_setup(
    multi_source_test_configs=SQLITE_ONLY,
    base_data=MISSING_AND_UNEXPECTED_DF,
    comparison_data=MISSING_AND_UNEXPECTED_DF,
)
def test_column_ordering(
    multi_source_batch: MultiSourceBatch,
    base_query: str,
    comparison_query: str,
    missing_rows: list[dict[str, Any]],
    unexpected_rows: list[dict[str, Any]],
) -> None:
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            base_query=base_query,
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=comparison_query.replace(
                "{source_table}", multi_source_batch.comparison_table_name
            ),
        )
    )

    assert result.result["details"] == {
        "unexpected_rows": unexpected_rows,
        "missing_rows": missing_rows,
    }


TOO_BIG_DATA = pd.DataFrame({"a": list(range(0, 500)), "b": list(range(100, 600))})


@multi_source_batch_setup(
    multi_source_test_configs=SQLITE_ONLY,
    base_data=TOO_BIG_DATA,
    comparison_data=TOO_BIG_DATA,
)
def test_expect_query_results_to_match_comparison_limit(multi_source_batch: MultiSourceBatch):
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            base_query="SELECT * FROM {batch} ORDER BY a",
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=f"SELECT * FROM {multi_source_batch.comparison_table_name} ORDER BY a",
        )
    )
    assert result.success
    assert result.result["unexpected_count"] == 0

    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            base_query="SELECT * FROM {batch} ORDER BY a",
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=f"SELECT * FROM {multi_source_batch.comparison_table_name} "
            "ORDER BY a DESC",
        )
    )
    assert not result.success
    assert result.result["unexpected_count"] == MAX_RESULT_RECORDS


@multi_source_batch_setup(
    multi_source_test_configs=ALL_COMPARISON_TO_BASE_SOURCES,
    base_data=BASE_DATA,
    comparison_data=COMPARISON_DATA,
)
def test_expect_query_results_to_match_comparison_error(multi_source_batch: MultiSourceBatch):
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            base_query="SELECT b FROM {batch}",
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query="SELECT invalid_column FROM "
            f"{multi_source_batch.comparison_table_name}",
        )
    )
    assert not result.success
    assert list(result.exception_info.values())[0]["raised_exception"]


DATA_WITH_MANY_COLUMNS = pd.DataFrame({ch: [1, 2, 3] for ch in "abcdefgh"})
OTHER_DATA_WITH_MANY_COLUMNS = pd.DataFrame({ch: [4, 5, 6] for ch in "abcdefgh"})


@multi_source_batch_setup(
    multi_source_test_configs=SQLITE_ONLY,
    base_data=DATA_WITH_MANY_COLUMNS,
    comparison_data=DATA_WITH_MANY_COLUMNS,
)
def test_rendering_no_differences(multi_source_batch: MultiSourceBatch):
    """NOTE: the queries here use kinda weird ordering to ensure that our output table
    actually reflects the right order.
    """
    source_table = multi_source_batch.comparison_table_name
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            base_query="SELECT e, a, d, g, b, e FROM {batch} ORDER BY e",
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=f"SELECT g, d, g, c, e, a  FROM {source_table} ORDER BY g",
        )
    )
    result.render()

    assert result.rendered_content == []


@multi_source_batch_setup(
    multi_source_test_configs=SQLITE_ONLY,
    base_data=DATA_WITH_MANY_COLUMNS,
    comparison_data=OTHER_DATA_WITH_MANY_COLUMNS,
)
def test_rendering_with_missing_and_unexpected(multi_source_batch: MultiSourceBatch):
    """NOTE: the queries here use kinda weird ordering to ensure that our output table
    actually reflects the right order.
    """
    source_table = multi_source_batch.comparison_table_name
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            base_query="SELECT a, b, c, d FROM {batch} ORDER BY e",
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=f"SELECT e, f, g, h  FROM {source_table} ORDER BY g",
        )
    )
    result.render()

    assert result.rendered_content == [
        RenderedAtomicContent(
            name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
            value=RenderedAtomicValue(
                template="Unexpected rows found in current table: $row_count",
                params={
                    "row_count": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "value": 3,
                    }
                },
                header_row=[
                    RendererTableValue(
                        schema=RendererSchema(type=RendererValueType.STRING),
                        value=col_name,
                    )
                    for col_name in ["a", "b", "c", "d"]
                ],
                table=[
                    [
                        RendererTableValue(
                            schema=RendererSchema(type=RendererValueType.NUMBER),
                            value=value,
                        )
                        for _ in ["a", "b", "c", "d"]
                    ]
                    for value in [1, 2, 3]
                ],
            ),
            value_type="TableType",
        ),
        RenderedAtomicContent(
            name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
            value=RenderedAtomicValue(
                template="Expected rows not found in current table: $row_count",
                params={
                    "row_count": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "value": 3,
                    }
                },
                header_row=[
                    RendererTableValue(
                        schema=RendererSchema(type=RendererValueType.STRING),
                        value=col_name,
                    )
                    for col_name in ["e", "f", "g", "h"]
                ],
                table=[
                    [
                        RendererTableValue(
                            schema=RendererSchema(type=RendererValueType.NUMBER),
                            value=value,
                        )
                        for _ in ["e", "f", "g", "h"]
                    ]
                    for value in [4, 5, 6]
                ],
            ),
            value_type="TableType",
        ),
    ]


@multi_source_batch_setup(
    multi_source_test_configs=SQLITE_ONLY,
    comparison_data=pd.DataFrame({"foo": [1, 2, 3, 3]}),
    base_data=pd.DataFrame({"bar": [1, 4, 5, 5]}),
)
def test_rendering_with_one_column(multi_source_batch: MultiSourceBatch):
    source_table = multi_source_batch.comparison_table_name
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=f"SELECT foo FROM {source_table}",
            base_query="SELECT bar FROM {batch}",
        )
    )
    result.render()

    assert result.rendered_content == [
        RenderedAtomicContent(
            name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
            value_type="StringValueType",
            value=RenderedAtomicValue(
                schema={"type": "com.superconductive.rendered.string"},
                meta_notes={"format": MetaNotesFormat.STRING, "content": []},
                template="$ov__0 $ov__1 $ov__2 $exp__0 $exp__1 $exp__2",
                params={
                    "expected_value": {
                        "schema": RendererSchema(type=RendererValueType.ARRAY),
                        "value": [2, 3, 3],
                    },
                    "observed_value": {
                        "schema": RendererSchema(type=RendererValueType.ARRAY),
                        "value": [4, 5, 5],
                    },
                    "exp__0": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "render_state": ObservedValueRenderState.MISSING,
                        "value": 2,
                    },
                    "exp__1": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "render_state": ObservedValueRenderState.MISSING,
                        "value": 3,
                    },
                    "exp__2": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "render_state": ObservedValueRenderState.MISSING,
                        "value": 3,
                    },
                    "ov__0": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "render_state": ObservedValueRenderState.UNEXPECTED,
                        "value": 4,
                    },
                    "ov__1": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "render_state": ObservedValueRenderState.UNEXPECTED,
                        "value": 5,
                    },
                    "ov__2": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "render_state": ObservedValueRenderState.UNEXPECTED,
                        "value": 5,
                    },
                },
            ),
        )
    ]


@multi_source_batch_setup(
    multi_source_test_configs=SQLITE_ONLY,
    comparison_data=pd.DataFrame({"foo": [1]}),
    base_data=pd.DataFrame({"bar": [2]}),
)
def test_rendering_with_one_value(multi_source_batch: MultiSourceBatch):
    source_table = multi_source_batch.comparison_table_name
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=f"SELECT foo FROM {source_table}",
            base_query="SELECT bar FROM {batch}",
        )
    )
    result.render()

    assert result.rendered_content == [
        RenderedAtomicContent(
            name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
            value_type="StringValueType",
            value=RenderedAtomicValue(
                schema={"type": "com.superconductive.rendered.string"},
                template="Observed value: $base_value",
                params={
                    "base_value": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "value": 2,
                    },
                },
            ),
        ),
        RenderedAtomicContent(
            name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
            value_type="StringValueType",
            value=RenderedAtomicValue(
                schema={"type": "com.superconductive.rendered.string"},
                template="Expected value: $comparison_value",
                params={
                    "comparison_value": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "value": 1,
                    },
                },
            ),
        ),
    ]


@multi_source_batch_setup(
    multi_source_test_configs=SQLITE_ONLY,
    comparison_data=pd.DataFrame({"foo": [1, 2, 3]}),
    base_data=pd.DataFrame({"bar": []}),
)
def test_rendering_only_missing_rows_single_column(multi_source_batch: MultiSourceBatch):
    """Test rendering when only missing_rows exist with single column."""
    source_table = multi_source_batch.comparison_table_name
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=f"SELECT foo FROM {source_table}",
            base_query="SELECT bar FROM {batch}",
        )
    )
    result.render()

    assert result.rendered_content == [
        RenderedAtomicContent(
            name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
            value_type="StringValueType",
            value=RenderedAtomicValue(
                schema={"type": "com.superconductive.rendered.string"},
                meta_notes={"format": MetaNotesFormat.STRING, "content": []},
                template="$exp__0 $exp__1 $exp__2",
                params={
                    "expected_value": {
                        "schema": RendererSchema(type=RendererValueType.ARRAY),
                        "value": [1, 2, 3],
                    },
                    "observed_value": {
                        "schema": RendererSchema(type=RendererValueType.ARRAY),
                        "value": [],
                    },
                    "exp__0": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "render_state": ObservedValueRenderState.MISSING,
                        "value": 1,
                    },
                    "exp__1": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "render_state": ObservedValueRenderState.MISSING,
                        "value": 2,
                    },
                    "exp__2": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "render_state": ObservedValueRenderState.MISSING,
                        "value": 3,
                    },
                },
            ),
        )
    ]


@multi_source_batch_setup(
    multi_source_test_configs=SQLITE_ONLY,
    comparison_data=pd.DataFrame({"foo": []}),
    base_data=pd.DataFrame({"bar": [1, 2, 3]}),
)
def test_rendering_only_unexpected_rows_single_column(multi_source_batch: MultiSourceBatch):
    """Test rendering when only unexpected_rows exist with single column."""
    source_table = multi_source_batch.comparison_table_name
    result = multi_source_batch.base_batch.validate(
        gxe.ExpectQueryResultsToMatchComparison(
            comparison_data_source_name=multi_source_batch.comparison_data_source_name,
            comparison_query=f"SELECT foo FROM {source_table}",
            base_query="SELECT bar FROM {batch}",
        )
    )
    result.render()

    assert result.rendered_content == [
        RenderedAtomicContent(
            name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
            value_type="StringValueType",
            value=RenderedAtomicValue(
                schema={"type": "com.superconductive.rendered.string"},
                meta_notes={"format": MetaNotesFormat.STRING, "content": []},
                template="$ov__0 $ov__1 $ov__2",
                params={
                    "expected_value": {
                        "schema": RendererSchema(type=RendererValueType.ARRAY),
                        "value": [],
                    },
                    "observed_value": {
                        "schema": RendererSchema(type=RendererValueType.ARRAY),
                        "value": [1, 2, 3],
                    },
                    "ov__0": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "render_state": ObservedValueRenderState.UNEXPECTED,
                        "value": 1,
                    },
                    "ov__1": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "render_state": ObservedValueRenderState.UNEXPECTED,
                        "value": 2,
                    },
                    "ov__2": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "render_state": ObservedValueRenderState.UNEXPECTED,
                        "value": 3,
                    },
                },
            ),
        )
    ]


@pytest.mark.postgresql
def test_unhashable_data_types():
    df = pd.DataFrame({"json_data": [{"foo": "bar"}]})
    context = gx.get_context(mode="ephemeral")
    batch_setup_a = PostgresBatchTestSetup(
        config=PostgreSQLDatasourceTestConfig(
            column_types={"json_data": postgresql_dialect.postgresqltypes.JSONB}
        ),
        data=df,
        extra_data={},
        context=context,
    )
    batch_setup_b = PostgresBatchTestSetup(
        config=PostgreSQLDatasourceTestConfig(
            column_types={"json_data": postgresql_dialect.postgresqltypes.JSONB}
        ),
        data=df,
        extra_data={},
        context=context,
    )

    with (
        batch_setup_a.batch_test_context() as batch_a,
        batch_setup_b.asset_test_context() as asset_b,
    ):
        data_source_name = asset_b.datasource.name
        source_table = asset_b.table_name
        expectation = gxe.ExpectQueryResultsToMatchComparison(
            comparison_data_source_name=data_source_name,
            comparison_query=f"SELECT * FROM {source_table}",
            base_query="SELECT * FROM {batch}",
        )

        result = batch_a.validate(expectation)

        assert result.exception_info["exception_message"] == "Unhashable column: json_data"


uuid_a = uuid.uuid4()
uuid_b = uuid.uuid4()
uuid_c = uuid.uuid4()
uuid_d = uuid.uuid4()


@pytest.mark.postgresql
def test_rendering_table_with_multiple_uuid():
    context = gx.get_context(mode="ephemeral")
    batch_setup_a = PostgresBatchTestSetup(
        config=PostgreSQLDatasourceTestConfig(
            column_types={"id": postgresql_dialect.postgresqltypes.UUID}
        ),
        data=pd.DataFrame({"name": ["a", "b"], "id": [uuid_a, uuid_b]}),
        extra_data={},
        context=context,
    )
    batch_setup_b = PostgresBatchTestSetup(
        config=PostgreSQLDatasourceTestConfig(
            column_types={"id": postgresql_dialect.postgresqltypes.UUID}
        ),
        data=pd.DataFrame({"name": ["a", "b"], "id": [uuid_c, uuid_d]}),
        extra_data={},
        context=context,
    )

    with (
        batch_setup_a.batch_test_context() as batch_a,
        batch_setup_b.asset_test_context() as asset_b,
    ):
        data_source_name = asset_b.datasource.name
        source_table = asset_b.table_name

        result = batch_a.validate(
            gxe.ExpectQueryResultsToMatchComparison(
                comparison_data_source_name=data_source_name,
                comparison_query=f"SELECT name, id FROM {source_table}",
                base_query="SELECT name, id FROM {batch}",
            )
        )
        result.render()

        assert result.rendered_content == [
            _create_table_rendered_atomic_content(
                template="Unexpected rows found in current table: $row_count",
                params={
                    "row_count": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "value": 2,
                    }
                },
                col_names=["name", "id"],
                col_types=[RendererValueType.STRING, RendererValueType.STRING],
                rows=[
                    ["a", uuid_a],
                    ["b", uuid_b],
                ],
            ),
            _create_table_rendered_atomic_content(
                template="Expected rows not found in current table: $row_count",
                params={
                    "row_count": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "value": 2,
                    }
                },
                col_names=["name", "id"],
                col_types=[RendererValueType.STRING, RendererValueType.STRING],
                rows=[
                    ["a", uuid_c],
                    ["b", uuid_d],
                ],
            ),
        ]


def _create_table_rendered_atomic_content(
    template: str,
    params: dict[str, Any],
    col_names: list[str],
    col_types: list[RendererValueType],
    rows: list[list[str]],
) -> RenderedAtomicContent:
    return RenderedAtomicContent(
        name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
        value=RenderedAtomicValue(
            template=template,
            params=params,
            header_row=[
                RendererTableValue(
                    schema=RendererSchema(type=RendererValueType.STRING),
                    value=col_name,
                )
                for col_name in col_names
            ],
            table=[
                [
                    RendererTableValue(
                        schema=RendererSchema(type=col_type),
                        value=cell_value,
                    )
                    for cell_value, col_type in zip(row, col_types)
                ]
                for row in rows
            ],
        ),
        value_type="TableType",
    )


@multi_source_batch_setup(
    multi_source_test_configs=ALL_COMPARISON_TO_BASE_SOURCES,
    base_data=BASE_DATA,
    comparison_data=COMPARISON_DATA,
)
def test_success_with_suite_param_base_query_(
    multi_source_batch: MultiSourceBatch,
) -> None:
    suite_param_key = "test_expect_query_results_to_match_comparison"

    expectation = gxe.ExpectQueryResultsToMatchComparison(
        base_query={"$PARAMETER": suite_param_key},
        comparison_data_source_name=multi_source_batch.comparison_data_source_name,
        comparison_query=f"SELECT a, b FROM {multi_source_batch.comparison_table_name} ORDER BY a, b",  # noqa: E501
    )

    result = multi_source_batch.base_batch.validate(
        expectation,
        expectation_parameters={suite_param_key: "SELECT a, b FROM {batch} ORDER BY a, b"},
    )
    assert result.success


@multi_source_batch_setup(
    multi_source_test_configs=ALL_COMPARISON_TO_BASE_SOURCES,
    base_data=BASE_DATA,
    comparison_data=COMPARISON_DATA,
)
def test_success_with_suite_param_comparison_query_(
    multi_source_batch: MultiSourceBatch,
) -> None:
    suite_param_key = "test_expect_query_results_to_match_comparison"

    expectation = gxe.ExpectQueryResultsToMatchComparison(
        base_query="SELECT a, b FROM {batch} ORDER BY a, b",
        comparison_data_source_name=multi_source_batch.comparison_data_source_name,
        comparison_query={"$PARAMETER": suite_param_key},
    )

    result = multi_source_batch.base_batch.validate(
        expectation,
        expectation_parameters={
            suite_param_key: f"SELECT a, b FROM {multi_source_batch.comparison_table_name} ORDER BY a, b"  # noqa: E501
        },
    )
    assert result.success



from __future__ import annotations

from typing import Literal, Union

from great_expectations._docs_decorators import public_api
from great_expectations.datasource.fluent.config_str import ConfigStr
from great_expectations.datasource.fluent.sql_datasource import SQLDatasource


@public_api
class BigQueryDatasource(SQLDatasource):
    """Adds a bigquery datasource to the data context.
    Args:
        name: The name of this big query datasource.
        connection_string: The connection string used to connect to the database.
            For example: "bigquery://<gcp_project_name>/<bigquery_dataset>"
        assets: An optional dictionary whose keys are TableAsset or QueryAsset names and whose
            values are TableAsset or QueryAsset objects.
    """

    type: Literal["bigquery"] = "bigquery"  # type: ignore[assignment]
    connection_string: Union[ConfigStr, str]



from typing import ClassVar

from great_expectations.expectations.metrics.query_metrics.query_row_count.query_row_count import (
    QueryRowCount,
)


class UnexpectedRowsQueryRowCount(QueryRowCount):
    metric_name = "unexpected_rows_query.row_count"
    value_keys = ("unexpected_rows_query",)

    query_param_name: ClassVar[str] = "unexpected_rows_query"



import os
import shutil
from collections import OrderedDict

import pytest
from pytest_mock import MockerFixture

import great_expectations as gx
from great_expectations.core.config_provider import _ConfigurationSubstitutor
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context import get_context
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.data_context.types.base import (
    DataContextConfig,
    DataContextConfigSchema,
)
from great_expectations.data_context.util import file_relative_path
from great_expectations.exceptions import InvalidConfigError, MissingConfigVariableError
from tests.data_context.conftest import create_data_context_files

yaml = YAMLHandler()

dataContextConfigSchema = DataContextConfigSchema()


@pytest.fixture
def empty_data_context_with_config_variables(monkeypatch, empty_data_context):
    monkeypatch.setenv("FOO", "BAR")
    monkeypatch.setenv("REPLACE_ME_ESCAPED_ENV", "ive_been_$--replaced")
    root_dir = empty_data_context.root_directory
    ge_config_path = file_relative_path(
        __file__,
        "../test_fixtures/great_expectations_basic_with_variables.yml",
    )
    shutil.copy(
        ge_config_path,
        os.path.join(root_dir, FileDataContext.GX_YML),  # noqa: PTH118 # FIXME CoP
    )
    config_variables_path = file_relative_path(
        __file__,
        "../test_fixtures/config_variables.yml",
    )
    shutil.copy(
        config_variables_path,
        os.path.join(root_dir, "uncommitted"),  # noqa: PTH118 # FIXME CoP
    )
    return get_context(context_root_dir=root_dir)


@pytest.mark.filesystem
def test_config_variables_on_context_without_config_variables_filepath_configured(
    data_context_without_config_variables_filepath_configured,
):
    # test the behavior on a context that does not config_variables_filepath (the location of
    # the file with config variables values) configured.

    context = data_context_without_config_variables_filepath_configured

    # an attempt to save a config variable should raise an exception

    with pytest.raises(InvalidConfigError) as exc:
        context.save_config_variable("var_name_1", {"n1": "v1"})
    assert "'config_variables_file_path' property is not found in config" in exc.value.message


@pytest.mark.filesystem
def test_substituted_config_variables_not_written_to_file(tmp_path_factory):
    # this test uses a great_expectations.yml with almost all values replaced
    # with substitution variables

    project_path = str(tmp_path_factory.mktemp("data_context"))
    context_path = os.path.join(project_path, FileDataContext.GX_DIR)  # noqa: PTH118 # FIXME CoP
    asset_config_path = os.path.join(context_path, "expectations")  # noqa: PTH118 # FIXME CoP

    create_data_context_files(
        context_path,
        asset_config_path,
        ge_config_fixture_filename="great_expectations_v013_basic_with_exhaustive_variables.yml",
        config_variables_fixture_filename="config_variables_exhaustive.yml",
    )

    # load ge config fixture for expected
    path_to_yml = "../test_fixtures/great_expectations_v013_basic_with_exhaustive_variables.yml"
    path_to_yml = file_relative_path(__file__, path_to_yml)
    with open(path_to_yml) as data:
        config_commented_map_from_yaml = yaml.load(data)
    expected_config = DataContextConfig.from_commented_map(config_commented_map_from_yaml)
    expected_config_commented_map = dataContextConfigSchema.dump(expected_config)

    # instantiate data_context twice to go through cycle of loading config from file then saving
    context = get_context(context_root_dir=context_path)
    context._save_project_config()
    context_config_commented_map = dataContextConfigSchema.dump(
        get_context(context_root_dir=context_path)._project_config
    )

    assert context_config_commented_map == expected_config_commented_map


@pytest.mark.unit
def test_substitute_config_variable():
    config_substitutor = _ConfigurationSubstitutor()
    config_variables_dict = {
        "arg0": "val_of_arg_0",
        "arg2": {"v1": 2},
        "aRg3": "val_of_aRg_3",
        "ARG4": "val_of_ARG_4",
    }
    assert (
        config_substitutor.substitute_config_variable("abc${arg0}", config_variables_dict)
        == "abcval_of_arg_0"
    )
    assert (
        config_substitutor.substitute_config_variable("abc$arg0", config_variables_dict)
        == "abcval_of_arg_0"
    )
    assert (
        config_substitutor.substitute_config_variable("${arg0}", config_variables_dict)
        == "val_of_arg_0"
    )
    assert (
        config_substitutor.substitute_config_variable("hhhhhhh", config_variables_dict) == "hhhhhhh"
    )
    with pytest.raises(MissingConfigVariableError) as exc:
        config_substitutor.substitute_config_variable(
            "abc${arg1} def${foo}", config_variables_dict
        )  # does NOT equal "abc${arg1}"
    assert "Unable to find a match for config substitution variable: `arg1`." in exc.value.message
    assert (
        config_substitutor.substitute_config_variable("${arg2}", config_variables_dict)
        == config_variables_dict["arg2"]
    )
    assert exc.value.missing_config_variable == "arg1"

    # Null cases
    assert config_substitutor.substitute_config_variable("", config_variables_dict) == ""
    assert config_substitutor.substitute_config_variable(None, config_variables_dict) is None

    # Test with mixed case
    assert (
        config_substitutor.substitute_config_variable(
            "prefix_${aRg3}_suffix", config_variables_dict
        )
        == "prefix_val_of_aRg_3_suffix"
    )
    assert (
        config_substitutor.substitute_config_variable("${aRg3}", config_variables_dict)
        == "val_of_aRg_3"
    )
    # Test with upper case
    assert (
        config_substitutor.substitute_config_variable("prefix_$ARG4/suffix", config_variables_dict)
        == "prefix_val_of_ARG_4/suffix"
    )
    assert (
        config_substitutor.substitute_config_variable("$ARG4", config_variables_dict)
        == "val_of_ARG_4"
    )

    # Test with multiple substitutions
    assert (
        config_substitutor.substitute_config_variable("prefix${arg0}$aRg3", config_variables_dict)
        == "prefixval_of_arg_0val_of_aRg_3"
    )

    # Escaped `$` (don't substitute, but return un-escaped string)
    assert (
        config_substitutor.substitute_config_variable(r"abc\${arg0}\$aRg3", config_variables_dict)
        == "abc${arg0}$aRg3"
    )

    # Multiple configurations together
    assert (
        config_substitutor.substitute_config_variable(
            r"prefix$ARG4.$arg0/$aRg3:${ARG4}/\$dontsub${arg0}:${aRg3}.suffix",
            config_variables_dict,
        )
        == "prefixval_of_ARG_4.val_of_arg_0/val_of_aRg_3:val_of_ARG_4/$dontsubval_of_arg_0:val_of_aRg_3.suffix"  # noqa: E501 # FIXME CoP
    )


@pytest.mark.unit
def test_escape_all_config_variables(empty_data_context_with_config_variables):
    """
    Make sure that all types of input to escape_all_config_variables are escaped properly: str, dict, OrderedDict, list
    Make sure that changing the escape string works as expected.
    """  # noqa: E501 # FIXME CoP
    context = empty_data_context_with_config_variables

    # str
    value_str = "pas$word1"
    escaped_value_str = r"pas\$word1"
    assert context.escape_all_config_variables(value=value_str) == escaped_value_str

    value_str2 = "pas$wor$d1$"
    escaped_value_str2 = r"pas\$wor\$d1\$"
    assert context.escape_all_config_variables(value=value_str2) == escaped_value_str2

    # dict
    value_dict = {
        "drivername": "postgresql",
        "host": os.getenv("GE_TEST_LOCAL_DB_HOSTNAME", "localhost"),
        "port": "5432",
        "username": "postgres",
        "password": "pass$word1",
        "database": "postgres",
    }
    escaped_value_dict = {
        "drivername": "postgresql",
        "host": os.getenv("GE_TEST_LOCAL_DB_HOSTNAME", "localhost"),
        "port": "5432",
        "username": "postgres",
        "password": r"pass\$word1",
        "database": "postgres",
    }
    assert context.escape_all_config_variables(value=value_dict) == escaped_value_dict

    # OrderedDict
    value_ordered_dict = OrderedDict(
        [
            ("UNCOMMITTED", "uncommitted"),
            ("docs_test_folder", "test$folder"),
            (
                "test_db",
                {
                    "drivername": "postgresql",
                    "host": "some_host",
                    "port": "5432",
                    "username": "postgres",
                    "password": "pa$sword1",
                    "database": "postgres",
                },
            ),
        ]
    )
    escaped_value_ordered_dict = OrderedDict(
        [
            ("UNCOMMITTED", "uncommitted"),
            ("docs_test_folder", r"test\$folder"),
            (
                "test_db",
                {
                    "drivername": "postgresql",
                    "host": "some_host",
                    "port": "5432",
                    "username": "postgres",
                    "password": r"pa\$sword1",
                    "database": "postgres",
                },
            ),
        ]
    )
    assert (
        context.escape_all_config_variables(value=value_ordered_dict) == escaped_value_ordered_dict
    )

    # list
    value_list = [
        "postgresql",
        os.getenv("GE_TEST_LOCAL_DB_HOSTNAME", "localhost"),
        "5432",
        "postgres",
        "pass$word1",
        "postgres",
    ]
    escaped_value_list = [
        "postgresql",
        os.getenv("GE_TEST_LOCAL_DB_HOSTNAME", "localhost"),
        "5432",
        "postgres",
        r"pass\$word1",
        "postgres",
    ]
    assert context.escape_all_config_variables(value=value_list) == escaped_value_list

    # Custom escape string
    value_str_custom_escape_string = "pas$word1"
    escaped_value_str_custom_escape_string = "pas@*&$word1"
    assert (
        context.escape_all_config_variables(
            value=value_str_custom_escape_string, dollar_sign_escape_string="@*&$"
        )
        == escaped_value_str_custom_escape_string
    )

    value_str_custom_escape_string2 = "pas$wor$d1$"
    escaped_value_str_custom_escape_string2 = "pas@*&$wor@*&$d1@*&$"
    assert (
        context.escape_all_config_variables(
            value=value_str_custom_escape_string2, dollar_sign_escape_string="@*&$"
        )
        == escaped_value_str_custom_escape_string2
    )


@pytest.mark.unit
def test_escape_all_config_variables_skip_substitution_vars(
    empty_data_context_with_config_variables,
):
    """
    What does this test and why?
    escape_all_config_variables(skip_if_substitution_variable=True/False) should function as documented.
    """  # noqa: E501 # FIXME CoP
    context = empty_data_context_with_config_variables

    # str
    value_str = "$VALUE_STR"
    escaped_value_str = r"\$VALUE_STR"
    assert (
        context.escape_all_config_variables(value=value_str, skip_if_substitution_variable=True)
        == value_str
    )
    assert (
        context.escape_all_config_variables(value=value_str, skip_if_substitution_variable=False)
        == escaped_value_str
    )

    value_str2 = "VALUE_$TR"
    escaped_value_str2 = r"VALUE_\$TR"
    assert (
        context.escape_all_config_variables(value=value_str2, skip_if_substitution_variable=True)
        == escaped_value_str2
    )
    assert (
        context.escape_all_config_variables(value=value_str2, skip_if_substitution_variable=False)
        == escaped_value_str2
    )

    multi_value_str = "${USER}:pas$word@${HOST}:${PORT}/${DATABASE}"
    escaped_multi_value_str = r"\${USER}:pas\$word@\${HOST}:\${PORT}/\${DATABASE}"
    assert (
        context.escape_all_config_variables(
            value=multi_value_str, skip_if_substitution_variable=True
        )
        == multi_value_str
    )
    assert (
        context.escape_all_config_variables(
            value=multi_value_str, skip_if_substitution_variable=False
        )
        == escaped_multi_value_str
    )

    multi_value_str2 = "$USER:pas$word@$HOST:${PORT}/${DATABASE}"
    escaped_multi_value_str2 = r"\$USER:pas\$word@\$HOST:\${PORT}/\${DATABASE}"
    assert (
        context.escape_all_config_variables(
            value=multi_value_str2, skip_if_substitution_variable=True
        )
        == multi_value_str2
    )
    assert (
        context.escape_all_config_variables(
            value=multi_value_str2, skip_if_substitution_variable=False
        )
        == escaped_multi_value_str2
    )

    multi_value_str3 = "USER:pas$word@$HOST:${PORT}/${DATABASE}"
    escaped_multi_value_str3 = r"USER:pas\$word@\$HOST:\${PORT}/\${DATABASE}"
    assert (
        context.escape_all_config_variables(
            value=multi_value_str3, skip_if_substitution_variable=True
        )
        == escaped_multi_value_str3
    )
    assert (
        context.escape_all_config_variables(
            value=multi_value_str3, skip_if_substitution_variable=False
        )
        == escaped_multi_value_str3
    )

    # dict
    value_dict = {
        "drivername": "postgresql",
        "host": "${HOST}",
        "port": "5432",
        "username": "postgres",
        "password": "pass$word1",
        "database": "$postgres",
        "sub_dict": {
            "test_val_no_escaping": "test_val",
            "test_val_escaping": "te$t_val",
            "test_val_substitution": "$test_val",
            "test_val_substitution_braces": "${test_val}",
        },
    }
    escaped_value_dict = {
        "drivername": "postgresql",
        "host": r"\${HOST}",
        "port": "5432",
        "username": "postgres",
        "password": r"pass\$word1",
        "database": r"\$postgres",
        "sub_dict": {
            "test_val_no_escaping": "test_val",
            "test_val_escaping": r"te\$t_val",
            "test_val_substitution": r"\$test_val",
            "test_val_substitution_braces": r"\${test_val}",
        },
    }
    escaped_value_dict_skip_substitution_variables = {
        "drivername": "postgresql",
        "host": "${HOST}",
        "port": "5432",
        "username": "postgres",
        "password": r"pass\$word1",
        "database": "$postgres",
        "sub_dict": {
            "test_val_no_escaping": "test_val",
            "test_val_escaping": r"te\$t_val",
            "test_val_substitution": "$test_val",
            "test_val_substitution_braces": "${test_val}",
        },
    }
    assert (
        context.escape_all_config_variables(value=value_dict, skip_if_substitution_variable=False)
        == escaped_value_dict
    )
    assert (
        context.escape_all_config_variables(value=value_dict, skip_if_substitution_variable=True)
        == escaped_value_dict_skip_substitution_variables
    )

    # OrderedDict
    value_ordered_dict = OrderedDict(
        [
            ("UNCOMMITTED", "uncommitted"),
            ("docs_test_folder", "test$folder"),
            (
                "test_db",
                {
                    "drivername": "$postgresql",
                    "host": "some_host",
                    "port": "5432",
                    "username": "${USERNAME}",
                    "password": "pa$sword1",
                    "database": "postgres",
                },
            ),
        ]
    )
    escaped_value_ordered_dict = OrderedDict(
        [
            ("UNCOMMITTED", "uncommitted"),
            ("docs_test_folder", r"test\$folder"),
            (
                "test_db",
                {
                    "drivername": r"\$postgresql",
                    "host": "some_host",
                    "port": "5432",
                    "username": r"\${USERNAME}",
                    "password": r"pa\$sword1",
                    "database": "postgres",
                },
            ),
        ]
    )
    escaped_value_ordered_dict_skip_substitution_variables = OrderedDict(
        [
            ("UNCOMMITTED", "uncommitted"),
            ("docs_test_folder", r"test\$folder"),
            (
                "test_db",
                {
                    "drivername": "$postgresql",
                    "host": "some_host",
                    "port": "5432",
                    "username": "${USERNAME}",
                    "password": r"pa\$sword1",
                    "database": "postgres",
                },
            ),
        ]
    )
    assert (
        context.escape_all_config_variables(
            value=value_ordered_dict, skip_if_substitution_variable=False
        )
        == escaped_value_ordered_dict
    )
    assert (
        context.escape_all_config_variables(
            value=value_ordered_dict, skip_if_substitution_variable=True
        )
        == escaped_value_ordered_dict_skip_substitution_variables
    )

    # list
    value_list = [
        "postgresql",
        os.getenv("GE_TEST_LOCAL_DB_HOSTNAME", "localhost"),
        "5432",
        "$postgres",
        "pass$word1",
        "${POSTGRES}",
    ]
    escaped_value_list = [
        "postgresql",
        os.getenv("GE_TEST_LOCAL_DB_HOSTNAME", "localhost"),
        "5432",
        r"\$postgres",
        r"pass\$word1",
        r"\${POSTGRES}",
    ]
    escaped_value_list_skip_substitution_variables = [
        "postgresql",
        os.getenv("GE_TEST_LOCAL_DB_HOSTNAME", "localhost"),
        "5432",
        "$postgres",
        r"pass\$word1",
        "${POSTGRES}",
    ]
    assert (
        context.escape_all_config_variables(value=value_list, skip_if_substitution_variable=False)
        == escaped_value_list
    )
    assert (
        context.escape_all_config_variables(value=value_list, skip_if_substitution_variable=True)
        == escaped_value_list_skip_substitution_variables
    )


@pytest.mark.filesystem
def test_create_data_context_and_config_vars_in_code(
    tmp_path_factory, monkeypatch, mocker: MockerFixture
):
    """
    What does this test and why?
    Creating a DataContext via .create(), then using .save_config_variable() to save a variable that will eventually be substituted (e.g. ${SOME_VAR}) should result in the proper escaping of $.
    This is in response to issue #2196
    """  # noqa: E501 # FIXME CoP

    project_path = str(tmp_path_factory.mktemp("data_context"))
    context = gx.get_context(
        mode="file",
        project_root_dir=project_path,
    )

    CONFIG_VARS = {
        "DB_HOST": "${DB_HOST_FROM_ENV_VAR}",
        "DB_NAME": "DB_NAME",
        "DB_USER": "DB_USER",
        "DB_PWD": "pas$word",
    }
    for k, v in CONFIG_VARS.items():
        context.save_config_variable(k, v)

    config_vars_file_contents = context.config_variables

    # Add escaping for DB_PWD since it is not of the form ${SOMEVAR} or $SOMEVAR
    CONFIG_VARS_WITH_ESCAPING = CONFIG_VARS.copy()
    CONFIG_VARS_WITH_ESCAPING["DB_PWD"] = r"pas\$word"

    # Ensure all config vars saved are in the config_variables.yml file
    # and that escaping was added for "pas$word" -> "pas\$word"
    assert all(
        item in config_vars_file_contents.items() for item in CONFIG_VARS_WITH_ESCAPING.items()
    )
    assert not all(item in config_vars_file_contents.items() for item in CONFIG_VARS.items())

    # Add env var for substitution
    DB_HOST_FROM_ENV_VAR_VALUE = "localhost"
    monkeypatch.setenv("DB_HOST_FROM_ENV_VAR", DB_HOST_FROM_ENV_VAR_VALUE)

    connection_string = "postgresql://${DB_USER}:${DB_PWD}@${DB_HOST}:65432/${DB_NAME}"
    mock_create_engine = mocker.Mock()
    mock_create_engine = mocker.patch("sqlalchemy.create_engine")

    context.data_sources.add_postgres("test_datasource", connection_string=connection_string)

    mock_create_engine.assert_called_once_with(
        "postgresql://DB_USER:pas$word@localhost:65432/DB_NAME"
    )

    # Ensure skip_if_substitution_variable=False works as documented
    context.save_config_variable("escaped", "$SOME_VAR", skip_if_substitution_variable=False)
    context.save_config_variable(
        "escaped_curly", "${SOME_VAR}", skip_if_substitution_variable=False
    )

    config_vars_file_contents = context.config_variables

    assert config_vars_file_contents["escaped"] == r"\$SOME_VAR"
    assert config_vars_file_contents["escaped_curly"] == r"\${SOME_VAR}"



from __future__ import annotations

from contextlib import contextmanager
from typing import Any, Dict, List


class MockSaInspector:
    def get_columns(self) -> List[Dict[str, Any]]:  # type: ignore[empty-body] # FIXME CoP
        ...

    def get_schema_names(self) -> List[str]:  # type: ignore[empty-body] # FIXME CoP
        ...

    def get_table_names(self, schema: str | None): ...

    def has_table(self, table_name: str, schema: str | None) -> bool:  # type: ignore[empty-body] # FIXME CoP
        ...


class Dialect:
    def __init__(self, dialect: str):
        self.name = dialect


class _MockConnection:
    def __init__(self, dialect: Dialect):
        self.dialect = dialect

    @contextmanager
    def begin(self):
        yield _MockConnection(self.dialect)

    def execute(self, statement):
        self.statement = statement


class MockSaEngine:
    def __init__(self, dialect: Dialect):
        self.dialect = dialect

    @contextmanager
    def begin(self):
        yield _MockConnection(self.dialect)

    @contextmanager
    def connect(self):
        """A contextmanager that yields a _MockConnection"""
        yield _MockConnection(self.dialect)

    def execute(self, *args, **kwargs):
        """This method is needed because currently we sometimes use a
        connection in place of an engine.

        When this is cleaned up we should remove this method.
        """
        pass



"""
The real meat of these tests lives in the test_definitions directory for individual expectations
"""

import pytest

from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.render.renderer.content_block import ExpectationStringRenderer


@pytest.mark.unit
def test_expectation_string_renderer_styling():
    renderer = ExpectationStringRenderer()
    result = renderer.render(
        ExpectationConfiguration(
            type="expect_column_values_to_be_unique",
            kwargs={"column": "Name"},
        )
    )
    assert len(result) == 1
    assert result[0].string_template["template"] == "$column values must be unique."

    result = renderer.render(
        ExpectationConfiguration(
            type="expect_column_values_to_be_unique",
            kwargs={"column": "Name", "mostly": 0.3},
        )
    )
    assert len(result) == 1
    template = result[0].string_template
    assert (
        template["template"] == "$column values must be unique, at least $mostly_pct % of the time."
    )
    assert template["params"]["mostly_pct"] == "30"

    result = renderer.render(
        ExpectationConfiguration(
            type="expect_column_values_to_be_unique",
            kwargs={"column": "Name", "mostly": 0.32345},
        )
    )
    assert len(result) == 1
    template = result[0].string_template
    assert (
        template["template"] == "$column values must be unique, at least $mostly_pct % of the time."
    )
    assert template["params"]["mostly_pct"] == "32.345"



from __future__ import annotations

from great_expectations.compatibility.not_imported import NotImported

DATABRICKS_CONNECT_NOT_IMPORTED = NotImported(
    "databricks-connect is not installed, please 'pip install databricks-connect'"
)

# The following types are modeled after the following documentation that is part
# of the databricks package.
# tldr: SQLAlchemy application should (mostly) "just work" with Databricks,
# other than the exceptions below
# https://github.com/databricks/databricks-sql-python/blob/main/src/databricks/sqlalchemy/README.sqlalchemy.md

try:
    from databricks.sqlalchemy._types import (
        TIMESTAMP_NTZ as TIMESTAMP_NTZ,  # noqa: PLC0414, RUF100 # FIXME CoP
    )
except (ImportError, AttributeError):
    TIMESTAMP_NTZ = DATABRICKS_CONNECT_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP

try:
    from databricks.sqlalchemy._types import (
        DatabricksStringType as STRING,  # noqa: PLC0414, RUF100 # FIXME CoP
    )
except (ImportError, AttributeError):
    STRING = DATABRICKS_CONNECT_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP

try:
    from databricks.sqlalchemy._types import (
        TIMESTAMP as TIMESTAMP,  # noqa: PLC0414, RUF100 # FIXME CoP
    )
except (ImportError, AttributeError):
    TIMESTAMP = DATABRICKS_CONNECT_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP

try:
    from databricks.sqlalchemy._types import TINYINT as TINYINT  # noqa: PLC0414, RUF100 # FIXME CoP
except (ImportError, AttributeError):
    TINYINT = DATABRICKS_CONNECT_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP


class DATABRICKS_TYPES:
    """Namespace for Databricks dialect types"""

    TIMESTAMP_NTZ = TIMESTAMP_NTZ
    STRING = STRING
    TINYINT = TINYINT
    TIMESTAMP = TIMESTAMP
